import math

import matplotlib.pyplot as plt

from Constants import CMaps, PltOpts
from cartesian_plot_functions import save_fig


def polplt_line(theta_ax, r_data, r_label, title, res_dir, sub_dir=None,
                fname='file_name', r_max=None):
    """Polar 2D line plot"""
    ax, fig = setup_polplt()
    ax.plot(theta_ax, r_data, 'k-', color='0', markersize='10')
    format_polplt(ax, r_max, r_label, title)
    return save_fig(res_dir, sub_dir, fname)


def polplt_line_line(theta_ax1, r_data1, theta_ax2, r_data2, r_label, title,
                     label1, label2, res_dir, sub_dir=None,
                     fname='file_name', r_max=None):
    """Polar 2D line plot with 2 lines"""
    ax, fig = setup_polplt()
    ax.plot(theta_ax1, r_data1, 'k-', label=label1)
    ax.plot(theta_ax2, r_data2, 'k--', label=label2)
    format_polplt(ax, r_max, r_label, title)
    return save_fig(res_dir, sub_dir, fname)


def polplt_scatt_line(theta_ax1, r_data1, theta_ax2, r_data2, r_label, title,
                      res_dir, sub_dir=None, fname='file_name',
                      r_max=None):
    """Polar 2D scatter and line plot"""
    ax, fig = setup_polplt()
    ax.scatter(theta_ax1, r_data1, s=PltOpts.msize.value, c=r_data1,
               cmap=CMaps.markers.value)
    ax.plot(theta_ax2, r_data2, 'k--', color='0')
    format_polplt(ax, r_max, r_label, title)
    return save_fig(res_dir, sub_dir, fname)


def setup_polplt():
    """Generic figure setup for 2D polar plot"""
    fig = plt.figure(facecolor='w')
    ax = plt.subplot(111, polar=True)
    ax.grid(True)
    return ax, fig


def format_polplt(ax, r_max, r_label, title):
    """Format 2D polar plot labels and figure layout"""
    ax.set_theta_zero_location("S")
    ax.set_rlabel_position(112.5)
    label_position = ax.get_rlabel_position()
    if r_max is None:
        ax.text(math.radians(label_position + 15), ax.get_rmax() / 2, r_label,
                rotation=22.5, ha='center', va='center')
    else:
        ax.text(math.radians(label_position + 15), r_max / 2, r_label,
                rotation=22.5, ha='center', va='center')
    plt.title(title, y=1.08)
    if r_max is not None:
        ax.set_rmax(r_max)



from abc import ABCMeta, abstractmethod

from BluPrintTriboSys import TriboSys


class RollBear(TriboSys):
    # TODO: add class attribute describtion
    """

    Attributes:
        wheels: An integer representing the number of wheels the vehicle has.

    """

    __metaclass__ = ABCMeta

    def __init__(self, name, number_rollers, roller, inner_ring, outer_ring,
                 global_force, init_force):
        super().__init__(name, global_force, init_force)
        self.roller = roller
        self.ring1 = inner_ring
        self.ring2 = outer_ring
        self.num_rollers = number_rollers
        self.rot_vel1 = 0
        self.rot_vel2 = 0

        self.mean_diam = None
        self.eff_rot_vel = None
        self.eff_omega = None
        self.eff_omega_cage = None

        self.footpr_vel = None
        self.actual_raceway_vel_roller = None

    @abstractmethod
    def calc_load_distribution(self, ui=None, res_dir=None):
        """Calculate the load distribution within the tribosystem"""
        pass

    @abstractmethod
    def calc_contact_pressure(self, ui=None, res_dir=None):
        """Calculate the contact pressure for each contact in the tribosystem"""
        pass

    @abstractmethod
    def calc_kinematics(self, rot_velocity, rot_velocity2, ui=None,
                        res_dir=None):
        """Calculate the kinematics of the tribosystem"""
        pass

    @abstractmethod
    def calc_pv(self, ui=None, res_dir=None):
        """"Calculate the product of the local maximum pressure and the
        relative velocity between ring and roller"""
        pass

    @abstractmethod
    def calc_e_akin(self, ui=None, res_dir=None):
        """"Calculate the kinetic friction energy accumulation"""
        pass

    @abstractmethod
    def plot_it(self, ui=None, res_dir=None):
        """Plot relevant calculation results"""
        pass

    @abstractmethod
    def generate_latex_output(self, calc_spec_tex_file_handle, simulation,
                              ui=None, res_dir=None):
        pass

    @abstractmethod
    def generate_latex_figures(self, ui=None, res_dir=None):
        pass


class AxialRollBear(RollBear):
    """Global bearing data"""

    def __init__(self, name, number_rollers, roller, shaft_ring, housing_ring,
                 global_force, mean_diameter):
        super().__init__(name, number_rollers, roller, shaft_ring, housing_ring,
                         global_force,
                         global_force / number_rollers)

        self.mean_diameter = mean_diameter
        self.pv = None

    @abstractmethod
    def calc_load_distribution(self, ui=None, res_dir=None):
        """Calculate the load distribution within the tribosystem"""
        pass

    @abstractmethod
    def calc_contact_pressure(self, ui=None, res_dir=None):
        """Calculate the contact pressure for each contact in the tribosystem"""
        pass

    @abstractmethod
    def calc_kinematics(self, rot_velocity, rot_velocity2, ui=None,
                        res_dir=None):
        """Calculate the kinematics of the tribosystem"""
        pass

    @abstractmethod
    def calc_pv(self, ui=None, res_dir=None):
        """"Calculate the product of the local maximum pressure and the relative
         velocity between ring and roller"""
        pass

    @abstractmethod
    def calc_e_akin(self, ui=None, res_dir=None):
        """"Calculate the kinetic friction energy accumulation"""
        pass

    @abstractmethod
    def plot_it(self, ui=None, res_dir=None):
        """Plot relevant calculation results"""
        pass

    @abstractmethod
    def generate_latex_output(self, calc_spec_tex_file_handle, simulation,
                              ui=None, res_dir=None):
        pass

    @abstractmethod
    def generate_latex_figures(self, ui=None, res_dir=None):
        pass


class RadialRollBear(RollBear):
    """Global bearing data"""

    def __init__(self, name, number_rollers, roller, inner_ring, outer_ring,
                 global_force, radial_clearance, res_pol,
                 path_roller_slip):
        super().__init__(name, number_rollers, roller, inner_ring, outer_ring,
                         global_force, global_force)

        self.mean_diameter = (self.ring1.diameter - self.ring2.diameter) / 2
        self.rad_clear = radial_clearance
        self.pv_ring1 = None
        self.pv_ring2 = None
        self.path_roller_slip = path_roller_slip
        self.res_pol = res_pol
        self.pol_ax = None
        self.dd_phi = None
        self.d_phi = None
        self.num_roller_pos = None
        self.phi_mat = None
        self.phi = None

    @abstractmethod
    def calc_load_distribution(self, ui=None, res_dir=None):
        """Calculate the load distribution within the tribosystem"""
        pass

    @abstractmethod
    def calc_contact_pressure(self, ui=None, res_dir=None):
        """Calculate the contact pressure for each contact in the tribosystem"""
        pass

    @abstractmethod
    def calc_kinematics(self, rot_velocity, rot_velocity2, ui=None,
                        res_dir=None):
        """Calculate the kinematics of the tribosystem"""
        pass

    @abstractmethod
    def calc_pv(self, ui=None, res_dir=None):
        """"Calculate the product of the local maximum pressure and the relative
        velocity between ring and roller"""
        pass

    @abstractmethod
    def calc_e_akin(self, ui=None, res_dir=None):
        """"Calculate the kinetic friction energy accumulation"""
        pass

    @abstractmethod
    def calc_e_akin(self, ui=None, res_dir=None):
        """"Calculate the kinetic friction energy accumulation"""
        pass

    @abstractmethod
    def plot_it(self, ui=None, res_dir=None):
        """Plot relevant calculation results"""
        pass

    @abstractmethod
    def generate_latex_output(self, calc_spec_tex_file_handle, simulation,
                              ui=None, res_dir=None):
        """Generate LaTeX output"""
        pass

    @abstractmethod
    def generate_latex_figures(self, ui=None, res_dir=None):
        """Generate LaTeX figures"""
        pass



# -*- coding: utf-8 -*-

"""

This module contains functions related to lubricants and lubrication.

"""


from math import log10 as lg
from math import pi, e

import numpy as np


def kin2dyn(kin, density):
    """

    Convert from kinematic to dynamic viscosity.

    Parameters
    ----------
    kin: ndarray, scalar
        The kinematic viscosity of the lubricant.
    density: ndarray, scalar
        The density of the lubricant.

    Returns
    -------
    dyn: ndarray, scalar
        The dynamic viscosity of the lubricant.

    """
    dyn = kin / density
    return dyn


def dyn2kin(dyn, density):
    """

    Convert from dynamic to kinematic viscosity.

    Parameters
    ----------
    dyn: ndarray, scalar
        The dynamic viscosity of the lubricant.
    density: ndarray, scalar
        The density of the lubricant.

    Returns
    -------
    kin: ndarray, scalar
        The kinematic viscosity of the lubricant.

    """
    kin = dyn * density
    return kin


def __zedwalther(kin):
    """

    Calculate the z-parameter for the Walther equation (ASTM D341).

    Parameters
    ----------
    kin: scalar
        The kinematic viscosity of the lubricant.

    Returns
    -------
    zed: scalar
        The z-parameter.

    """
    zed = kin + 0.7 + 10 ** (-1.47 - 1.84 * kin - 0.51 * kin ** 2)
    return zed


def __nuwalther(zed):
    """

    Calculate the kinematic viscosity for the Walther equation (ASTM D341).

    Parameters
    ----------
    zed: scalar
        The z-parameter of the Walther equation.

    Returns
    -------
    kin: scalar
        The kinematic viscosity.

    """
    kin = (zed - 0.7) - 10 ** (-0.7487 - 3.295 * (zed - 0.7) +
                               0.6119 * (zed - 0.7) ** 2 - 0.3193 *
                               (zed - 0.7) ** 3)
    return kin


def walther(temp_1, nu_1, temp_2, nu_2, temp_3):
    """

    Calculate the kinematic viscosity at temperature `temp_3` based on the
    kinematic viscosities at temperatures `temp_1` and `temp_2`. The
    implementation follows standard ASTM D341.

    Parameters
    ----------
    temp_1: scalar
        The temperature in :math:`^{\\circ}\\text{C}` that corresponds to the
        kinematic viscosity :code:`nu_1`.
    nu_1: scalar
        The kinematic viscosity in cSt at temperature :code:`temp_1`.
    temp_2: scalar
        The temperature in :math:`^{\\circ}\\text{C}` that corresponds to the
        kinematic viscosity :code:`nu_2`.
    nu_2: scalar
        The kinematic viscosity in cSt at temperature :code:`temp_2`.
    temp_3: ndarray
        The temperature in :math:`^{\\circ}\\text{C}` for which to calculate
        the kinematic viscosity.

    Returns
    -------
    nu_3: ndarray
        The kinematic viscosity in cSt at temperature :code:`temp_3`.

    """
    abs_zero = -273.15
    viscs = [nu_1, nu_2]
    thetas = [temp_1 - abs_zero, temp_2 - abs_zero, temp_3 - abs_zero]

    zed = [__zedwalther(nu) for nu in viscs]
    const_a = (lg(lg(zed[0])) - lg(lg(zed[1])) *
               lg(thetas[0]) / lg(thetas[1])) / \
              (1 - lg(thetas[0]) / lg(thetas[1]))
    const_b = (const_a - lg(lg(zed[1]))) / lg(thetas[1])
    nu_3 = __nuwalther(10 ** 10 ** (const_a - const_b * np.log10(thetas[2])))
    return nu_3


def trheomflat(eta, gap_height, omega, r_a):
    """

    Calculate the torque that occurs during a viscosity measurement using a
    plate-on-plate rheometer.

    Parameters
    ----------
    eta: scalar
        The dynamic viscosity of the fluid.
    gap_height: scalar
        The gap height between the stationary and the rotating plate.
    omega: scalar
        The angular velocity of the rotating plate (in radians per second).
    r_a: scalar
        The outer diameter of the rotating plate.

    Returns
    -------
    torque: scalar
        The torque required to rotate the rotating plate.

    """
    torque = eta * pi * omega * r_a ** 4 / (2 * gap_height)
    return torque


def viscrheomflat(torque, gap_height, omega, r_a):
    """

    Calculate the dynamic viscosity of a fluid based on the measurement torque
    in a plate-on-plate rheometer.

    Parameters
    ----------
    torque: scalar
        The torque required to rotate the rotating plate.
    gap_height: scalar
        The gap height between the stationary and the rotating plate.
    omega: scalar
        The angular velocity of the rotating plate (in radians per second).
    r_a: scalar
        The outer diameter of the rotating plate.

    Returns
    -------
    eta: scalar
        The dynamic viscosity of the fluid.

    """
    eta = 2 * gap_height * torque / (pi * omega * r_a ** 4)
    return eta


def trheomcone(eta, alpha, omega, r_a):
    """

    Calculate the torque that occurs during a viscosity measurement using a
    cone-on-plate rheometer.

    Parameters
    ----------
    eta: scalar
        The dynamic viscosity of the fluid.
    alpha: scalar
        The cone angle in radians.
    omega: scalar
        The angular velocity of the rotating plate (in radians per second).
    r_a: scalar
        The outer diameter of the rotating plate.

    Returns
    -------
    torque: scalar
        The torque required to rotate the rotating plate.

    """
    torque = 2 * pi * eta * omega * r_a ** 3 / (3 * alpha)
    return torque


def viscrheomcone(torque, alpha, omega, r_a):
    """

    Calculate the dynamic viscosity of a fluid based on the measurement torque
    in a cone-on-plate rheometer.

    Parameters
    ----------
    torque: scalar
        The torque required to rotate the rotating plate.
    alpha: scalar
        The cone angle in radians.
    omega: scalar
        The angular velocity of the rotating plate (in radians per second).
    r_a: scalar
        The outer diameter of the rotating plate.

    Returns
    -------
    eta: scalar
        The dynamic viscosity of the fluid.

    """
    eta = 3 * torque * alpha / (2 * pi * omega * r_a ** 3)
    return eta


def barus(eta_0, alpha_p, pressure):
    """

    Calculate the dynamic viscosity at a given pressure based on the dynamic
    viscosity at atmospheric pressure and the pressure-viscosity coefficient.

    Parameters
    ----------
    eta_0: scalar
        The dynamic viscosity at atmospheric pressure.
    alpha_p: scalar
        The pressure-viscosity coefficient of the fluid.
    pressure: scalar
        The pressure at which to calculate the dynamic viscosity.

    Returns
    -------
    eta_p: scalar
        The dynamic viscosity at pressure `pressure`.

    """
    eta_p = eta_0 * e ** (alpha_p * pressure)
    return eta_p



# -*- coding: utf-8 -*-

"""

Relevant constants for tribology research. Use, for example, as follows:

.. code-block:: python

   # get young's modulus of steel in units of MPa
   import tribology as tr
   youngs_modulus_steel = tr.YoungsMod.STEEL.value

"""

from enum import Enum


class RadBall(Enum):
    """

    Scalar constants for ball radii in units of mm. Available values are:

    :TQInch: 3/4 inch
    :HInch: 1/2 inch
    :Inch: 1 inch

    """
    TQInch = 25.4 * 3 / 8
    HInch = 25.4 / 4
    Inch = 25.4 / 2


class YoungsMod(Enum):
    """

    Scalar constants for Young's modulus in units of MPa. Available values are:

    :STEEL:
    :GLASS:
    :SiN:

    """
    STEEL = 210000
    GLASS = 70000
    SiN = 315000


class PoissonRatio(Enum):
    """

    Scalar constants for unitless Poisson ratios. Available values are:

    :STEEL:
    :GLASS:
    :SiN:

    """
    STEEL = 0.3
    GLASS = 0.22
    SiN = 0.26


class MatDens(Enum):
    """

    Scalar constants for material density in :math:`\\text{kg m}^{-3}`.
    Available values are:

    :STEEL:
    :GLASS:

    """
    STEEL = 7800
    SiN = 3200


class LubeDens(Enum):
    """

    Scalar constants for lubricant density in :math:`\\text{g ml}^{-1}`.
    Available values are:

    """
    MINERAL_OIL_GENERIC = 0.862


class LubeVisc(Enum):
    """

    Tuple containing kinematic viscosities (in unit of cSt) of lubricants
    at 40 and 100 :math:`^{\\circ}\\text{C}`. The first value in the tuple
    corresponds to the viscosity at 40 :math:`^{\\circ}\\text{C}`, the second
    to 100 :math:`^{\\circ}\\text{C}`. The **TEMPS** value contains the
    reference temperatures (40, 100). Available values are:

    :TEMPS: Tuple containing temperature values of 40 and 100

    """
    TEMPS = (40, 100)


class PressVisc(Enum):
    """

    Pressure-viscosity coefficients in :math:`\\text{Pa}^{-1}` at room
    temperature. Available values are:

    :ESTER_OIL_GENERIC:
    :MINERAL_OIL_GENERIC:

    """
    ESTER_OIL_GENERIC = 15 * 10 ** (-9)
    MINERAL_OIL_GENERIC = 30 * 10 ** (-9)



import gc
import os

import matplotlib.pyplot as plt
import numpy as np

from Constants import CMaps, PltOpts, SubDir, PrintOpts
from ContactBodies import ContactBody
from system_functions import make_directory, print_it, print_progress


def plt_profile_approx(res_dir, sub_dir):
    """If contact body profile was generated from txt file input, plot the
    approximated profile together with the profile loaded from the file"""
    for body in gc.get_objects():
        if isinstance(body, ContactBody):
            if body.type_profile == "File":
                fname = "profile-approximation-{}".format(body.name)
                plt_2d_scatt_line(body.file_x_axis, body.file_x_profile,
                                  body.x_axis, body.x_profile,
                                  body.x_label, body.z_label, fname, res_dir,
                                  sub_dir, fname)


def plt_surf_roughness(res_dir):
    """If contact body profile is rough, plot surface roughness"""
    for body in gc.get_objects():
        if isinstance(body, ContactBody):
            if body.roughness_mat is not None:
                fname = "surface-roughness-{}".format(body.name)
                print(fname)
                plt_3d(body.x_axis, body.y_axis, body.roughness_mat,
                       body.x_label, body.y_label,
                       'roughness height in mm',
                       'surface roughness {}'.format(body.name), res_dir,
                       sub_dir=SubDir.profiles.value, fname=fname)


def save_fig(res_dir, sub_dir, fname):
    """Save figure with name 'fname' to png format"""
    if sub_dir is None:
        fhandle = os.sep.join([res_dir, fname + '.png'])
    else:
        fhandle = os.sep.join(
            [make_directory(res_dir, sub_dir), fname + '.png'])
    plt.savefig(fhandle,
                bbox_inches='tight',
                dpi=PltOpts.res.value)
    plt.close()
    return fhandle


def plt_3d(x_ax, y_ax, z_data, x_label, y_label, z_label, title, res_dir,
           sub_dir=None, fname='file_name',
           z_lim=None, azim=None):
    """Generate 3D surface plot"""
    from mpl_toolkits.mplot3d import Axes3D
    fig = plt.figure()
    ax = Axes3D(fig)
    x_grid, y_grid = np.meshgrid(x_ax, y_ax, sparse=False, indexing='ij')
    ax.plot_surface(x_grid, y_grid, z_data,
                    rstride=1,
                    cstride=1,
                    cmap=CMaps.default.value,
                    linewidth=0.5,
                    antialiased=True,
                    edgecolors='k')
    ax.set_xlabel('\n' + x_label)
    ax.set_ylabel('\n' + y_label)
    ax.set_zlabel('\n\n' + z_label)
    if z_lim is not None:
        ax.set_zlim(z_lim)
    if azim is not None:
        ax.view_init(azim=azim)
    ax.w_xaxis.set_pane_color((1.0, 1.0, 1.0, 1.0))
    ax.w_yaxis.set_pane_color((1.0, 1.0, 1.0, 1.0))
    ax.w_zaxis.set_pane_color((1.0, 1.0, 1.0, 1.0))
    plt.title(title, y=1.08)
    return save_fig(res_dir, sub_dir, fname)


def plt_2d(x_ax, y_data, x_label, y_label, title, res_dir, sub_dir=None,
           fname='file_name'):
    """Generate 2D scatter plot"""
    plt.figure()
    plt.scatter(x_ax, y_data,
                s=PltOpts.msize.value,
                c=y_data,
                cmap=CMaps.markers.value,
                edgecolors='k')
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(title, y=1.08)
    return save_fig(res_dir, sub_dir, fname)


def plt_contact(body1, body2, profile_type, res_dir, sub_dir=None, fname=None):
    """Generate contact plot, showing two un-deformed contact bodies"""
    if fname is None:
        fname = '-'.join([body2.name, body1.name, 'contact', profile_type])
    if profile_type == PltOpts.DD.value:
        plt.figure()
        plt.scatter(body2.x_axis, body2.x_profile,
                    s=PltOpts.msize.value,
                    c=body2.x_profile,
                    cmap=CMaps.markers.value,
                    edgecolors='k')
        plt.scatter(body1.x_axis, -body1.x_profile,
                    s=PltOpts.msize.value,
                    c=body1.x_profile,
                    cmap=CMaps.markers.value,
                    edgecolors='k')
        plt.xlabel(body2.x_label)
        plt.ylabel(body2.z_label)
    elif profile_type == PltOpts.DDD.value:
        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        ax = Axes3D(fig)
        x_grid, y_grid = np.meshgrid(body2.x_axis, body2.y_axis, sparse=False,
                                     indexing='ij')
        ax.plot_surface(x_grid, y_grid, body2.profile,
                        rstride=1,
                        cstride=1,
                        cmap=CMaps.contacts.value,
                        linewidth=0.5,
                        antialiased=True)
        ax.plot_surface(x_grid, y_grid, -body1.profile,
                        rstride=1,
                        cstride=1,
                        cmap=CMaps.contacts.value,
                        linewidth=0.5,
                        antialiased=True)
        ax.set_xlabel('\n' + body2.x_label)
        ax.set_ylabel('\n' + body2.y_label)
        ax.set_zlabel('\n\n' + 'profile in mm')
        ax.w_xaxis.set_pane_color((1.0, 1.0, 1.0, 1.0))
        ax.w_yaxis.set_pane_color((1.0, 1.0, 1.0, 1.0))
        ax.w_zaxis.set_pane_color((1.0, 1.0, 1.0, 1.0))
    plt.title('contact {} and {}'.format(body1.name, body2.name), y=1.08)
    return save_fig(res_dir, sub_dir, fname)


def plt_profile(body, profile_type, res_dir, sub_dir=None, fname=None):
    """Plot contact body profile in 2D or 3D"""
    if fname is None:
        fname = body.name + '-profile-' + profile_type
    # 2D
    if profile_type == PltOpts.DD.value:
        plt_2d(body.x_axis, body.x_profile, body.x_label, body.y_label,
               (body.name + '-profile'), res_dir,
               sub_dir, fname)
    # 3D
    elif profile_type == PltOpts.DDD.value:
        plt_3d(body.x_axis, body.y_axis, body.profile, body.x_label,
               body.y_label, 'profile in mm',
               (body.name + '-profile'), res_dir, sub_dir, fname)


def plt_2d_scatt_line(x_ax1, y_data1, x_ax2, y_data2, x_label, y_label, title,
                      res_dir, sub_dir=None,
                      fname='file_name'):
    """Plot 2D scatter and line plot"""
    plt.figure()
    plt.plot(x_ax1, y_data1, '--', color='0')
    plt.scatter(x_ax2, y_data2, s=PltOpts.msize.value, c=y_data2,
                cmap=CMaps.markers.value, edgecolors='k')
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(title, y=1.08)
    return save_fig(res_dir, sub_dir, fname)


def plt_2d_2y_ax(x_ax1, y_data1, x_ax2, y_data2, x_label, title, leg1, leg2,
                 res_dir, sub_dir=None, fname='file_name'):
    """Plot 2D line plot with two y-axis"""
    fig, ax1 = plt.subplots()
    plot1 = ax1.plot(x_ax1, y_data1, '-', color='0', label=leg1)
    ax1.set_xlabel(x_label)
    ax1.set_ylabel(leg1)
    plt.title(title, y=1.08)
    ax2 = ax1.twinx()
    plot2 = ax2.plot(x_ax2, y_data2, '--', color='0', label=leg2)
    ax2.set_ylabel(leg2)
    plots = plot1 + plot2
    labels = [l.get_label() for l in plots]
    legend = ax1.legend(plots, labels, loc=4)
    legend.get_frame().set_facecolor('#FFFFFF')
    return save_fig(res_dir, sub_dir, fname)


def plt_energy_ring_on_ring(tribo_system, res_dir, sub_dir, fname):
    """Plot energy figures for ring-on-ring tribosystems. This is a legacy
    function and should be replaced to follow the more general plotting approach
    that other plot functions use"""
    fig, ax1 = plt.subplots()
    plot1 = ax1.plot(tribo_system.sun.x_axis, tribo_system.sun.e_akin, '-',
                     color='0.75', label='e_a,kin sun')
    plot2 = ax1.plot(tribo_system.planet.x_axis, tribo_system.planet.e_akin,
                     '--', color='0', label='e_a,kin planet')
    ax1.set_xlabel(tribo_system.sun.x_label)
    ax1.set_ylabel('e_akin in W $\mathregular{m^{-2}}$')
    plt.title('e_akin vs pv_rel', y=1.08)
    ax2 = ax1.twinx()
    plot3 = ax2.plot(tribo_system.sun.x_axis, tribo_system.pv, '-.',
                     color='0.75', label='pv_rel sun/planet')
    ax2.set_ylabel('pv_rel in W $\mathregular{m^{-2}}$')
    plots = plot1 + plot2 + plot3
    labels = [l.get_label() for l in plots]
    legend = ax1.legend(plots, labels, loc=4)
    legend.get_frame().set_facecolor('#FFFFFF')
    return save_fig(res_dir, sub_dir, fname)


def plt_cum_dist(data, x_label, res_dir, sub_dir, fname):
    """Plot 2D cumulative distribution of data"""
    values, base = np.histogram(data, bins=200)
    cumulative = np.cumsum(values)
    plt.plot(base[:-1], cumulative / max(cumulative), c='k')
    plt.xlim(0, np.amax(data))
    plt.xlabel(x_label)
    plt.ylabel('cumulative distribution')
    return save_fig(res_dir, sub_dir, fname)


def plt_profiles(res_dir):
    """plot profiles of all contact bodies using python's garbage collector to
    find all objects of type ContactBody"""
    print_it("plotting profiles", PrintOpts.lvl1.value)
    num_cbs = list(
        isinstance(obj, ContactBody) for obj in gc.get_objects()).count(True)
    counter = 0
    for body in gc.get_objects():
        if isinstance(body, ContactBody):
            if body.type_profile == "File":
                fname = "profile-approximation-{}".format(body.name)
                plt_2d_scatt_line(body.file_x_axis, body.file_x_profile,
                                  body.x_axis, body.x_profile,
                                  body.x_label, body.z_label, fname, res_dir,
                                  SubDir.profiles.value, fname)

            if body.roughness_mat is not None:
                fname = "surface-roughness-{}".format(body.name)
                plt_3d(body.x_axis, body.y_axis, body.roughness_mat,
                       body.x_label, body.y_label,
                       'roughness height in mm',
                       'surface roughness {}'.format(body.name), res_dir,
                       sub_dir=SubDir.profiles.value, fname=fname)

            plt_profile(body, PltOpts.DD.value, res_dir, SubDir.profiles.value)
            plt_profile(body, PltOpts.DDD.value, res_dir, SubDir.profiles.value)
            counter = print_progress(counter, num_cbs)



import os
import re
import copy

import numpy as np

from Constants import VarDicts, VarProps, OddEvenCheck, VarType, SimType, \
    Profiles
from system_functions import exit_program


def str_to_bool(s):
    """convert string representation of boolean to boolean"""
    if s == 'True':
        return True
    elif s == 'False':
        return False
    else:
        return s


def import_user_input(in_file):
    """read file input file line by line and try to extract user input.
    return dictionary with user input"""
    user_input = {}
    invalid_input = {}
    try:
        with open(in_file) as user_input_file:
            for line in user_input_file:
                # if line not empty
                if line.strip():
                    # if line is comment, ignore
                    if line.strip()[0] == '#':
                        continue
                    else:
                        # remove trailing comments and whitespace
                        value_pair = line.split('#')[0].strip(' ')
                        # split at '=' and remove whitespace
                        key = value_pair.split('=')[0].strip(' ')
                        value = value_pair.split('=')[1].strip('\n').strip(' ')
                        # value = value.strip("'").strip('"').strip("'")
                        user_input.update({key: value})
                        if value.startswith("'") or value.startswith('"'):
                            value = value[1:-1]
                            if re.search('[a-zA-Z]', value):
                                user_input.update({key: value})
                        elif value == 'True':
                            user_input.update({key: True})
                            if key == 'auto_print':
                                user_input.update({'AUTO_PRINT': 'True'})
                        elif value == 'False':
                            user_input.update({key: False})
                            if key == 'auto_print':
                                user_input.update({'AUTO_PRINT': 'False'})
                        else:
                            try:
                                user_input.update({key: float(value)})
                                if int(user_input[key]) == user_input[key]:
                                    user_input[key] = int(value)
                            except ValueError:
                                invalid_input.update({key: value})
        return user_input, invalid_input
    except FileNotFoundError:
        exit_program("can't find file {}".format(in_file))


def check_base_required(user_input):
    """checks if all basic (mandatory) variables are defined"""
    required_inputs = copy.deepcopy(VarDicts.required_user_input.value)
    for key, value in required_inputs.items():
        if key not in user_input:
            exit_program(
                "variable '{}' not defined in input file".format(key))
    check_vars(user_input)
    return required_inputs


def check_all_required(user_input, required_inputs):
    """checks if all calculation-specific (mandatory) variables are defined"""
    for key, value in required_inputs.items():
        if key not in user_input:
            exit_program(
                "variable '{}' not defined in input file".format(key))
    required_inputs.update(get_sim_specific_input(user_input)[0])
    check_vars(user_input)


def check_base_optional(user_input):
    """checks if all basic optional variables are defined properly"""
    to_check = {}
    optional_inputs = VarDicts.optional_user_input.value
    for key, value in user_input.items():
        if key in optional_inputs:
            to_check[key] = value
    check_vars(to_check)
    return optional_inputs


def check_all_optional(user_input, optional_inputs):
    """checks if all calculation-specific optional variables are defined
    properly"""
    to_check = {}
    optional_inputs.update(get_sim_specific_input(user_input)[1])
    for key, value in user_input.items():
        if key in optional_inputs:
            to_check[key] = value
    check_vars(to_check)


def check_sign(key, value, sign_var):
    """checks if the sign of a numeric variable is as expected"""
    if (isinstance(value, int) and not isinstance(value, bool)) is False \
            and isinstance(value, float) is False:
        exit_program(
            "variable '{}' with value '{}' should not be of type {}"
            .format(key, value, type(value)))
    if np.sign(value) != sign_var:
        exit_program("variable '{}' with value '{}' should be {} 0"
                     .format(key, value, {1.0: '<', -1.0: '>'}[np.sign(value)]))


def check_alpha(key, value):
    """Checks if the string has only alphanumeric characters, except for '_'
    and '-' """
    try:
        if value[0].isalnum() is False or value[-1].isalnum() is False:
            if value[0] == '-' or value[-1] == '-':
                exit_program(
                    "variable '" + key +
                    "' should not start or end with a hyphen '-' ")
            else:
                exit_program(
                    "variable '{}' with value '{}' should only contain "
                    "alphanumeric characters and hyphens '-'"
                    .format(key, value))
    except TypeError:
        exit_program(
            "variable '{}' with value '{}' should be of type string. strings "
            "need to be enclosed in quotes.".
            format(key, value))
    for character in value:
        if character.isalnum() is False and character != '-':
            exit_program(
                "variable '{}' with value '{}' should only contain "
                "alphanumeric characters and hyphens '-'".
                format(key, value))


def check_odd_even(key, value, should_be_odd_or_even=OddEvenCheck.Undefined):
    """Checks if the variable is odd or even"""
    if should_be_odd_or_even.value == OddEvenCheck.Odd.value or \
                    should_be_odd_or_even.value == OddEvenCheck.Even.value:
        if value % 2 == OddEvenCheck.Odd.value and \
                        should_be_odd_or_even.value != OddEvenCheck.Odd.value:
            exit_program("variable '" + value + "' should be an even number")
        elif value % 2 == OddEvenCheck.Even.value and \
                        should_be_odd_or_even.value != OddEvenCheck.Even.value:
            exit_program("variable '{}' should be an odd number".format(key))


def check_bool(key, value, isbool=True):
    """Checks if the variable is boolean"""
    if isinstance(value, bool) is not True and isbool is True:
        exit_program(
            "variable '{}' with value '{}' should be of type boolean".format(
                key, value))
    elif isinstance(value, bool) is True and isbool is False:
        exit_program(
            "variable '{}' with value '{}' (boolean) should not "
            "be of type boolean".format(key, value))


def check_int(key, value):
    """Checks if the variable is int"""
    if isinstance(value, int) is not True:
        exit_program(
            "variable '{}' with value '{}' should be of type int".format(key,
                                                                         value))


def check_vars(var_dict):
    """"Orchestrates variable checks"""
    for key, value in var_dict.items():
        if key in VarProps.bool_vars.value:
            check_bool(key, value, True)
        else:
            check_bool(key, value, False)
            if key in VarProps.positive_vars.value:
                check_sign(key, value, 1)
            if key in VarProps.negative_vars.value:
                check_sign(key, value, -1)
            if key in VarProps.odd_vars.value:
                check_odd_even(key, value, OddEvenCheck.Odd)
            if key in VarProps.alphanumeric_vars.value:
                check_alpha(key, value)
            if key in VarProps.int_vars.value:
                check_int(key, value)


def check_exists(key, user_input, else_quit=False):
    """Checks if the variable was defined by user, i.e., if it's in the
    global variables"""
    if key not in user_input and else_quit is True:
        exit_program("variable '{}' not defined".format(key))
    else:
        return user_input[key]


def get_sim_specific_input(user_input):
    """returns additional simulation-specific required and optional user
    input variables"""
    simulation_type = user_input['simulation_type']
    required_user_input = {}
    optional_user_input = {}
    if simulation_type == SimType.cyl_rol_bearing.value:
        required_user_input = {'length_cb1': VarType.real_number,
                               'e_cb3': VarType.real_number,
                               'ny_cb3': VarType.real_number,
                               'diameter_cb3': VarType.real_number,
                               'type_profile_cb1': VarType.string,
                               'type_profile_cb2': VarType.string,
                               'type_profile_cb3': VarType.string,
                               'number_rollers': VarType.integer,
                               'diameter_cb2': VarType.real_number}
        optional_user_input = {'radial_clearance': VarType.real_number,
                               'res_pol': VarType.integer,
                               'rot_velocity2': VarType.real_number,
                               'rms_roughness_cb3': VarType.real_number}
    elif simulation_type == SimType.deep_gro_ball_bearing.value:
        pass
    elif simulation_type == SimType.cyl_rol_thrust_bearing.value:
        required_user_input = {'length_cb1': VarType.real_number,
                               'e_cb3': VarType.real_number,
                               'ny_cb3': VarType.real_number,
                               'mean_diameter': VarType.real_number,
                               'type_profile_cb1': VarType.string,
                               'number_rollers': VarType.integer}
        optional_user_input = {'rot_velocity2': VarType.real_number}
    elif simulation_type == SimType.ball_on_disk.value:
        required_user_input = {'number_balls': VarType.integer,
                               'sliding_diameter': VarType.real_number,
                               'diameter_cb2': VarType.real_number,
                               'rot_velocity2': VarType.real_number}
    elif simulation_type == SimType.pin_on_disk.value:
        check_vars({'diameter_cb1': user_input['diameter_cb1'],
                    'type_profile_cb1': user_input['type_profile_cb1']})
        if user_input[
            'type_profile_cb1'] is Profiles.Circle.value and check_exists(
                'profile_radius_cb1') is True and \
                        user_input['diameter_cb1'] != 2 * user_input[
                    'profile_radius_cb1']:
            required_user_input.update({'length_cb1': VarType.real_number})
        required_user_input.update({'number_pins': VarType.integer,
                                    'sliding_diameter': VarType.real_number})
    elif simulation_type == SimType.four_ball.value:
        required_user_input = {'diameter_cb2': VarType.real_number}
    elif simulation_type == SimType.ball_on_three_plates.value:
        optional_user_input = {'plate_angle': VarType.real_number}
    elif simulation_type == SimType.ring_on_ring.value:
        required_user_input = {'length_cb1': VarType.real_number,
                               'type_profile_cb1': VarType.string,
                               'type_profile_cb2': VarType.string,
                               'number_planets': VarType.integer,
                               'rot_velocity2': VarType.real_number}
    else:
        exit_program(
            "'simulation_type' option '{}' not defined".format(simulation_type))
    return required_user_input, optional_user_input


def validate_file_path(path_variable):
    """Validates file path"""
    if os.path.isfile(path_variable) is False:
        exit_program("cannot find file '" + eval(path_variable) + "'"
                     "\nplease specify a valid file path (including file "
                     "extension) in variable '" + path_variable + "'" +
                     "\nthe file path needs to be defined relative to "
                     "'the input file'")


def validate_contact_body(user_input):
    """Checks if the user's contact body definition is consistent"""
    bodies_to_validate = {SimType.cyl_rol_bearing.value: ['cb1', 'cb2', 'cb3'],
                          SimType.cyl_rol_thrust_bearing.value: ['cb1'],
                          SimType.pin_on_disk.value: ['cb1'],
                          SimType.ring_on_ring.value: ['cb1', 'cb2']}
    for contact_body in bodies_to_validate.get(user_input['simulation_type'],
                                               []):
        type_profile = "type_profile_{}".format(contact_body)
        if user_input[type_profile] == Profiles.NoProfile.value:
            pass
        elif user_input[type_profile] == Profiles.ISO.value:
            check_vars({"length_" + contact_body: user_input[
                "length_" + contact_body]})
        elif user_input[type_profile] == Profiles.Circle.value:
            check_vars(
                {"length_" + contact_body: user_input["length_" + contact_body],
                 "profile_radius_" + contact_body: user_input[
                     "profile_radius_" + contact_body]})
        elif user_input[type_profile] == Profiles.File.value:
            validate_file_path(user_input["path_profile_" + contact_body])
        else:
            exit_program("'type_profile' option '{}' is undefined".format(
                user_input[type_profile]))


def get_user_input(in_file):
    """Orchestrates the process of user input handling"""
    user_input, invalid_input = import_user_input(in_file)
    required_inputs = check_base_required(user_input)
    optional_inputs = check_base_optional(user_input)
    spec_required_inputs, spec_optional_inputs = get_sim_specific_input(
        user_input)
    required_inputs.update(spec_required_inputs)
    optional_inputs.update(spec_optional_inputs)
    check_all_required(user_input, required_inputs)
    check_all_optional(user_input, optional_inputs)
    validate_contact_body(user_input)
    return user_input, invalid_input



# -*- coding: utf-8 -*-

"""

This module contains functions related to boundary element solvers for contact
mechanics calculations. Please see the :code:`/demo` directory for minimum
examples of how to combine the below functions to perform contact mechanics
calculations.

"""


import copy
from math import pi

import numexpr as ne
import numpy as np
import scipy.sparse.linalg as spla


def __secant(x_list, fx_list):
    """

    Applies secant method to find root *x* of function *f(x)*.
    If not enough (*x*, *f(x)*) value pairs are available to apply
    secant method, a new *x* value is guessed by slightly changing the
    initial *x* value

    Parameters
    ----------
    x_list: ndarray, list
        A list of known function inputs (x-values).
    fx_list: ndarray, list
        A list of known function outputs corresponding to the function inputs
        in ``x_list``.

    Returns
    -------
    x: scalar
        An estimate of the root of function f(x).

    """
    if fx_list[-1] != 0:
        if len(x_list) > 1 and \
                        fx_list[-1] != 0 and \
                        abs(fx_list[-1]) != abs(fx_list[-2]):
            x_0 = x_list[-2]
            x_1 = x_list[-1]
            fx_0 = fx_list[-2]
            fx_1 = fx_list[-1]
            slope = (fx_1 - fx_0) / (x_1 - x_0)
            return x_1 + (-fx_1 / slope)
        else:
            return x_list[0] * 0.9 + 0.0001
    else:
        return x_list[-1]


def __beinflumatgrid(axis):
    """

    Private helper method for influence matrix generation. Generates a grid
    that is required for the generation of an influence matrix.

    Parameters
    ----------
    axis: ndarray
        An axis of the elastic half space.

    Returns
    -------
    infl_mat_grid: ndarray
        A grid for influence matrix generation.

    """
    len_axis = len(axis)
    vec = np.zeros((1, len_axis))
    vec[0, :] = axis
    vertical_ax = np.zeros((len_axis, 1))
    vertical_ax[:, 0] = axis
    grid = np.repeat(vec, len_axis, axis=0)
    infl_mat_grid = np.absolute(np.subtract(grid, vertical_ax))
    return infl_mat_grid


def beinflumatred(infl_mat):
    """

    Calculate a reduced influence coefficient matrix from a complete influence
    coefficient matrix.

    Parameters
    ----------
    infl_mat: ndarray
        The complete influence coefficient matrix.

    Returns
    -------
    reduced_infl_mat: ndarray
        The reduced influence coefficient matrix. The matrix is square and of
        order ``n = np.shape(infl_mat)[0]`` :math:`\\times`
        ``np.shape(infl_mat)[1]``.

    """
    shape_mat = np.shape(infl_mat)
    len_mat = shape_mat[0] * shape_mat[1]
    reduced_infl_mat = np.zeros((len_mat, len_mat))
    counter = 0
    for i in range(0, shape_mat[0]):
        for j in range(0, shape_mat[1]):
            reduced_infl_mat[counter, :] = \
                np.reshape(infl_mat[i, j, :, :], len_mat)
            counter += 1
    return reduced_infl_mat


def beinflumat(x_axis, y_axis, e_eff):
    """

    Generate a complete influence coefficient matrix as required for boundary
    element contact mechanics calculations.

    Parameters
    ----------
    x_axis: ndarray
        The x-axis values of the coordinate grid.
    y_axis: ndarray
        The y-axis values of the coordinate grid.
    e_eff: scalar
        The effective modulus of the contact bodies.

    Returns
    -------
    infl_mat: ndarray
        The complete four-dimensional influence coefficient matrix of size
        ``len(x_axis)`` :math:`\\times` ``len(y_axis)`` :math:`\\times`
        ``len(x_axis)`` :math:`\\times` ``len(y_axis)``.

    """
    len_x = len(x_axis)
    len_y = len(y_axis)
    influence_matrix_complete = np.zeros((len_x, len_y, len_x, len_y))

    # generate coordinate grids
    a_factor = (x_axis[-1] - x_axis[0]) / (len_x - 1) / 2
    b_factor = (y_axis[-1] - y_axis[0]) / (len_y - 1) / 2
    x_grid = __beinflumatgrid(x_axis)
    y_grid = __beinflumatgrid(y_axis)

    # use numexpr to evaluate expressions
    xpa = ne.evaluate('x_grid + a_factor')
    xma = ne.evaluate('x_grid - a_factor')
    ypb = ne.evaluate('y_grid + b_factor')
    ymb = ne.evaluate('y_grid - b_factor')

    # calculate complete influence matrix
    for j in range(0, len_y):
        for j_prime in range(0, len_y):
            influence_matrix_complete[:, j, :, j_prime] =  \
                    (np.multiply(xpa, np.log(
                        np.divide(
                            ((ypb[j, j_prime]) +
                             np.sqrt(np.multiply((ypb[j, j_prime]),
                                                 (ypb[j, j_prime])) +
                                     np.multiply(xpa, xpa))),
                            ((ymb[j, j_prime]) +
                             np.sqrt(np.multiply((ymb[j, j_prime]),
                                                 (ymb[j, j_prime])) +
                                     np.multiply(xpa, xpa)))))) +
                     (ypb[j, j_prime]) * np.log(
                         np.divide(
                             (xpa +
                              np.sqrt(np.multiply((ypb[j, j_prime]),
                                                  (ypb[j, j_prime])) +
                                      np.multiply(xpa, xpa))),
                             (xma +
                              np.sqrt(np.multiply((ypb[j, j_prime]),
                                                  (ypb[j, j_prime])) +
                                      np.multiply(xma, xma))))) +
                     np.multiply(xma, np.log(
                         np.divide(
                             ((ymb[j, j_prime]) +
                              np.sqrt(np.multiply((ymb[j, j_prime]),
                                                  (ymb[j, j_prime])) +
                                      np.multiply(xma, xma))),
                             ((ypb[j, j_prime]) +
                              np.sqrt(np.multiply((ypb[j, j_prime]),
                                                  (ypb[j, j_prime])) +
                                      np.multiply(xma, xma)))))) +
                     (ymb[j, j_prime]) * np.log(
                         np.divide(
                             (xma +
                              np.sqrt(np.multiply((ymb[j, j_prime]),
                                                  (ymb[j, j_prime])) +
                                      np.multiply(xma, xma))),
                             (xpa +
                              np.sqrt(np.multiply((ymb[j, j_prime]),
                                                  (ymb[j, j_prime])) +
                                      np.multiply(xpa, xpa))))))

    infl_mat = influence_matrix_complete * 1 / e_eff * 2 / pi
    return infl_mat


def __begetd(profile, norm_disp):
    """

    Parameters
    ----------
    profile: ndarray
        The combined profiles of two contact bodies.
    norm_disp: scalar
        The normal elastic displacement (interference) of the two contact
        bodies.

    Returns
    -------
    disp: ndarray
        A displacement field containing the combined local elastic displacements
        of both contact bodies.

    """
    displ_field = np.subtract(np.ones(profile.shape) * norm_disp, profile)
    disp = np.reshape(displ_field, (profile.shape[0] * profile.shape[1]))
    return disp


def __solvepress(red_infl_mat, disp):
    """

    Solve for pressure distribution by removing negative pressure elements from
    the pressure matrix.

    Parameters
    ----------
    red_infl_mat: ndarray
        The recued influence matrix of the contact problem.
    disp: ndarray
        The (combined) displacement field of the contact problem.

    Returns
    -------
    pressure: ndarray
        The pressure field of the contact problem.

    """

    # find negative pressure arguments
    pressure = spla.gmres(red_infl_mat, disp)[0]
    p_index = np.zeros(len(pressure))
    negative_p = np.where(pressure < 0)[0]
    p_neg = copy.deepcopy(negative_p)

    while len(negative_p) > 0:
        pressure[p_neg] = 0
        p_index[p_neg] = 1
        u_new_reduced = np.delete(disp, [p_neg], axis=0)
        g_new_reduced = np.delete(red_infl_mat, [p_neg], axis=0)
        g_new_reduced = np.delete(g_new_reduced, [p_neg], axis=1)
        if pressure[np.where(p_index == 0)].size > 0:
            pressure[np.where(p_index == 0)] = \
                spla.gmres(g_new_reduced, u_new_reduced)[0]
        negative_p = np.where(pressure < 0)[0]
        p_neg = np.append(p_neg, negative_p)

    return pressure


def besolve(profile_1, profile_2, outer_force,
            red_infl_mat, delta_x, delta_y,
            norm_disp=0.1, max_offset=0.005):
    """

    Solve a system of linear equations to find the pressure and displacement
    distribution in a boundary element contact problem:

        :math:`[\\text{pressure}] = [\\text{reduced influence matrix}]^{-1}
        \\cdot [\\text{displacement}]`

    The start value for the relative normal displacement is defined by the
    ``norm_disp`` parameter. The default value corresponds to 0.1 units of
    length. If the start value of the normal displacement is close to the
    equilibrium displacement, the calculation time may be reduced significantly.

    The solver stops once inner and outer forces are in equilibrium. The maximum
    allowed discrepancy between inner and outer forces is defined by the
    ``max_offset`` parameter. The default value corresponds to 0.5 %.

    Parameters
    ----------
    profile_1: ndarray
        A matrix containing the profile heights of body 1.
    profile_2: ndarray
        A matrix containing the profile heights of body 2. The matrix must have
        the same size and grid spacing (in unit length) as ``profile_1``.
    outer_force: scalar
        The (outer) normal force acting on the contact.
    red_infl_mat: ndarray
        The reduced influence matrix calculated using the ``beinflumatred``
        method.
    delta_x: scalar
        The grid spacing of the profile matrices in unit length in x-direction.
    delta_y: scalar
        The grid spacing of the profile matrices in unit length in y-direction.
    norm_disp: scalar
        The initial normal elastic deformation in unit length used to start the
        solver.
    max_offset: scalar
        The maximum allowed percentage difference of inner and outer force (used
        to determine when solver has completed).

    Returns
    -------
    pressure: ndarray
        The contact pressure field in the contact.
    disp: ndarray
        The combined normal displacement field for contact body 1 and 2.
    inner_force: scalar
        The sum over all inner forces.
    norm_disp: scalar
        The normal elastic equilibrium displacement.

    """
    # initialise variables
    x_value = [0]
    fx_value = [outer_force]
    pressure = 0
    inner_force = 0
    profile = profile_1 + profile_2

    # while difference between inner forces and outer forces is significant
    while abs(inner_force - outer_force) > max_offset * outer_force:

        # update local displacements
        disp = __begetd(profile, norm_disp)
        pressure = __solvepress(red_infl_mat, disp)

        # calculate resulting force and adjust displacement for next loop
        pressure = np.reshape(pressure, (profile.shape[0], profile.shape[1]))
        inner_force = sum(sum(np.multiply(delta_x * delta_y, pressure)))
        x_value = np.append(x_value, [norm_disp])
        fx_value = np.append(fx_value, [inner_force - outer_force])
        norm_disp = __secant(x_value, fx_value)

    disp = __begetd(profile, norm_disp)
    return pressure, disp, inner_force, norm_disp



import math

import numpy as np

from numeric_methods import secant_method


def load_distr_cyl_rol_bear(phis, number_rollers, length_roller, res_x,
                            combined_profile, roller_axis, radial_clearance,
                            radial_load):
    """Caclulate the load distribution in a cylindrical roller bearing
    according to standard DIN 26281"""
    cl = 35948 * math.pow(length_roller, (8 / 9))
    cs = cl / res_x
    x_axis = abs(roller_axis)
    cos_phi = list((math.cos(phi) for phi in phis))

    delta_r = radial_clearance * 4 + 0.000000001
    delta_f = 20
    sum_re_ns = np.zeros(number_rollers)
    psi_j = np.zeros(number_rollers)
    delta_j = np.zeros(number_rollers)
    sum_mns = np.zeros(number_rollers)
    delta_jk = np.zeros((number_rollers, res_x))
    delta_re = np.zeros(number_rollers)
    x_value = []
    fx_value = []
    radial_load = radial_load or 0.000000001

    while abs(delta_f) > 0.0005 * radial_load:
        sum_zwk = 0
        delta_re = np.zeros(number_rollers)
        for z in range(number_rollers):
            sum_re_ns[z] = 0
            sum_mns[z] = 0
            delta_j[z] = delta_r * cos_phi[z] - radial_clearance / 2
            for n in range(0, res_x):
                delta_jk[z, n] = max((delta_j[z] - x_axis[n] * math.tan(
                    psi_j[z]) - 2 * combined_profile[n]), 0)
                sum_re_ns[z] += math.pow(delta_jk[z, n], (10 / 9))
            delta_re[z] = cos_phi[z] * sum_re_ns[z]
            sum_zwk += delta_re[z]
        delta_f = abs(radial_load - cs * sum_zwk)
        x_value = np.append(x_value, [delta_r])
        fx_value = np.append(fx_value, [delta_f])
        delta_r = secant_method(x_value, fx_value)

    roller_normal_forces = sum_re_ns * cs
    return roller_normal_forces, delta_re



# Option for variable 'simulation_type':
# 1: cylindrical roller bearing
# 2:
# 3: cylindrical roller thrust bearing
# 4: ball on disk (currently not fully supported)
# 5: pin on disk
# 6: 4 ball
# 7: ball on three plates
# 8: ring on ring

# global simulation setup
simulation_type = 3  # one of the above type numbers
simulation_name = 'ThrustRollerBearing'
auto_print = True  # True or False
auto_plot = True
auto_report = False  # reporting currently not supported, don't change

# global test setup / bearing information
tribo_system_name = 'ThrustBearing-81212'
number_rollers = 13
mean_diameter = 77.5

#  Contact Body 1 (CB1)
e_cb1 = 210000  # young's modulus in MPa
ny_cb1 = 0.3  # poisson number in MPa
diameter_cb1 = 10  # in mm
length_cb1 = 10  # in mm
type_profile_cb1 = 'File'  # 'None', 'ISO', 'Circle', 'File'
path_profile_cb1 = 'tribology/p3can/BearingProfiles/NU206-RE-1.txt'  # path to profile.txt file required if TypeProfile == 'File'
profile_radius_cb1 = 6.35  # input required if TypeProfile == 'Circle'
roller_slip = 'tribology/p3can/RollerSlip/testslip.txt'

#  Contact Body 2 (CB2)
e_cb2 = 210000  # young's modulus in MPa
ny_cb2 = 0.3  # poisson number in MPa

# Contact Body 3 (CB3)
e_cb3 = 210000  # young's modulus in MPa
ny_cb3 = 0.3  # poisson number in MPa

# Loads
global_force = 80000  # in N
rot_velocity1 = 300  # in rpm
rot_velocity2 = 0  # in rpm

# Mesh
res_x = 41  # data points along roller length
res_y = 39  # data points along roller width



# Option for variable 'simulation_type':
# 1: cylindrical roller bearing
# 2:
# 3: cylindrical roller thrust bearing
# 4: ball on disk (currently not fully supported)
# 5: pin on disk
# 6: 4 ball
# 7: ball on three plates
# 8: ring on ring

# global simulation setup
simulation_type = 7  # one of the above type numbers
simulation_name = 'BallOnThreePlates'
auto_print = True  # True or False
auto_plot = True
auto_report = False  # reporting currently not supported, don't change

# global test setup / bearing information
tribo_system_name = 'bla'
plate_angle = 45  # angle in degree

# Rotating Ball (CB1)
e_cb1 = 210000  # young's modulus in MPa
ny_cb1 = 0.3  # poisson number in MPa
diameter_cb1 = 12.7  # in mm

# Stationary Plates (CB2)
e_cb2 = 210000  # young's modulus in MPa
ny_cb2 = 0.3  # poisson number in MPa

# Loads
global_force = 50  # in N
rot_velocity1 = 50  # in rpm

# Mesh
res_x = 17  # data points along roller length
res_y = 15  # data points along roller width



import math


def hertz_displ(e1, e2, ny1, ny2, r1x, r1y, r2x, r2y, f):
    """Normal elastic displacement for arbitrary bodies according to hertz
    contact theory"""
    apb = 0.5 * (1 / r1x + 1 / r1y + 1 / r2x + 1 / r2y)
    bma = 0.5 * math.sqrt((1 / r1x - 1 / r1y) ** 2 + (1 / r2x - 1 / r2y) ** 2)
    ra = 1 / (apb - bma)
    rb = 1 / (apb + bma)
    rc = math.sqrt(ra * rb)

    f1 = 1 - math.pow(math.pow(ra / rb, 0.0602) - 1, 1.456)
    f2 = 1 - math.pow(math.pow(ra / rb, 0.0684) - 1, 1.531)
    ec = 1 / ((1 - ny1 ** 2) / e1 + (1 - ny2 ** 2) / e2)

    c = math.pow(3 * f * rc / (4 * ec), 1 / 3) * f1
    e = 1 - math.pow(rb / ra, 4 / 3)
    a = c * math.pow(1 - e ** 2, 1 / 4)
    b = c * math.pow(1 - e ** 2, 1 / 4)
    delta = a * b / rc * (f2 / f1)
    return delta


def get_diam_approx(x_axis, x_profile):
    """Approximate hertz contact radii based on actual contact body geometry"""
    x1 = x_axis[0]
    x2 = x_axis[math.floor(len(x_axis) / 2)]
    x3 = x_axis[-1]
    y1 = x_profile[0]
    y2 = x_profile[math.floor(len(x_axis) / 2)]
    y3 = x_profile[-1]

    a = math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)
    b = math.sqrt((x2 - x3) ** 2 + (y2 - y3) ** 2)
    c = math.sqrt((x3 - x1) ** 2 + (y3 - y1) ** 2)
    s = (a + b + c) / 2
    d = math.sqrt(s * (s - a) * (s - b) * (s - c))
    try:
        diam = 2 * a * b * c / (4 * d)
    except ZeroDivisionError:
        diam = 40000
    return diam


if __name__ == "__main__":
    hertz_displ(210000, 210000, 0.3, 0.3, 12.7 / 2, 12.7 / 2, 41039, 41039, 50)



from abc import ABCMeta, abstractmethod


class TriboSys(object):
    """All simulations rely on the definition of an object of type TriboSys,
    which contains all physical parameters describing the tribosystem,
    including contact bodies."""

    __metaclass__ = ABCMeta

    def __init__(self, name, global_force, init_force):
        self.name = name
        self.init_force = init_force
        self.global_force = global_force

    @abstractmethod
    def calc_load_distribution(self, ui=None, res_dir=None):
        """Calculate the load distribution within the tribosystem"""
        pass

    @abstractmethod
    def calc_contact_pressure(self, ui=None, res_dir=None):
        """Calculate the contact pressure for each contact in the tribosystem"""
        pass

    @abstractmethod
    def calc_kinematics(self, rot_velocity, rot_velocity2, ui=None,
                        res_dir=None):
        """Calculate the kinematics of the tribosystem"""
        pass

    @abstractmethod
    def calc_pv(self, ui=None, res_dir=None):
        """Calculate the kinematics of the tribosystem"""
        pass

    @abstractmethod
    def calc_e_akin(self, ui=None, res_dir=None):
        """Calculate the kinematics of the tribosystem"""
        pass

    @abstractmethod
    def plot_it(self, ui=None, res_dir=None):
        """Plot relevant calculation results"""
        pass

    @abstractmethod
    def generate_latex_output(self, calc_spec_tex_file_handle, simulation,
                              ui=None, res_dir=None):
        pass

    @abstractmethod
    def generate_latex_figures(self, ui=None, res_dir=None):
        pass



import math

from BluPrintAxialThrustBear import AxialThrustBear
from BluPrintBallOnThreePlates import BallOnThreePlates
from BluPrintCylRollBear import CylRollBear
from BluPrintFourBall import FourBall
from BluPrintPinOnDisk import PinOnDisk
from BluPrintRingOnRing import RingOnRing
from Constants import Profiles
from ContactBodies import Ring, Flat, Disk, Ball
from system_functions import exit_program, in_d


def setup_cyl_rol_bearing(ui):
    """Generate roller contact body (master) as well as inner and outer ring
    (slaves)"""
    roller = Ring('roller', ui['e_cb1'], ui['ny_cb1'], ui['diameter_cb1'],
                  ui['length_cb1'], ui['type_profile_cb1'],
                  rms_rough=in_d('rms_roughness_cb1', ui),
                  circle_radius=in_d('profile_radius_cb1', ui),
                  path_profile=in_d('path_profile_cb1', ui))
    roller.make_profile(ui['res_x'], ui['res_y'], ui['global_force'], 1)
    ring1 = Ring('inner-ring', ui['e_cb2'], ui['ny_cb2'], ui['diameter_cb2'],
                 ui['length_cb2'], ui['type_profile_cb2'],
                 rms_rough=in_d('rms_roughness_cb2', ui),
                 circle_radius=in_d('profile_radius_cb2', ui),
                 path_profile=in_d('path_profile_cb2', ui))
    ring1.make_slave_to(roller)
    ring2 = Ring('outer-ring', ui['e_cb3'], ui['ny_cb3'], -ui['diameter_cb3'],
                 ui['length_cb3'], ui['type_profile_cb3'],
                 rms_rough=in_d('rms_roughness_cb2', ui),
                 circle_radius=in_d('profile_radius_cb3', ui),
                 path_profile=in_d('path_profile_cb3', ui))
    ring2.make_slave_to(roller)
    tribo_system = CylRollBear(ui['number_rollers'], roller, ring1, ring2,
                               ui['global_force'], ui['radial_clearance'],
                               in_d('res_pol', ui), in_d('roller_slip', ui),
                               in_d('tribo_system_name', ui))
    return roller, ring1, ring2, tribo_system


def setup_deep_gro_ball_bearing(ui):
    """Not currently implemented"""
    exit_program("simulation_type '2' currently undefined")


def setup_cyl_rol_thrust_bearing(ui):
    """Generate roller contact body (master) as well as two flats (slave)"""
    roller = Ring('roller', ui['e_cb1'], ui['ny_cb1'], ui['diameter_cb1'],
                  ui['length_cb1'], ui['type_profile_cb1'],
                  rms_rough=in_d('rms_roughness_cb1', ui),
                  circle_radius=in_d('profile_radius_cb1', ui),
                  path_profile=in_d('path_profile_cb1', ui))
    roller.make_profile(ui['res_x'], ui['res_y'],
                        ui['global_force'] / ui['number_rollers'], 1)
    ring1 = Flat('shaft-washer', ui['e_cb2'], ui['ny_cb2'],
                 rms_rough=in_d('rms_roughness_cb2', ui))
    ring1.make_slave_to(roller)
    ring2 = Flat('housing-washer', ui['e_cb3'], ui['ny_cb3'],
                 rms_rough=in_d('rms_roughness_cb3', ui))
    ring2.make_slave_to(roller)
    tribo_system = AxialThrustBear(ui['number_rollers'], ui['mean_diameter'],
                                   roller, ring1, ring2,
                                   ui['global_force'],
                                   in_d('tribo_system_name', ui))
    return roller, ring1, ring2, tribo_system


def setup_ball_on_disk(ui):
    """Not currently implemented"""
    exit_program("simulation_type '4' currently not implemented completely.")
    """
    ball = Ball('pin', ui['e_cb1'], ui['ny_cb1'], ui['diameter_cb1'],
                rms_rough=in_d('rms_roughness_cb1', ui))
    ball.make_profile(ui['res_x'], ui['res_y'],
                      ui['global_force'] / ui['number_balls'])
    disk = Disk('disk', ui['e_cb2'], ui['ny_cb2'], diameter=ui['diameter_cb2'],
                rot_vel='rot_velocity2',
                rms_rough=in_d('rms_roughness_cb2', ui))
    disk.make_slave_to(ball)
    tribo_system = BallOnDisk(ui['number_balls'], ui['sliding_diameter'], ball,
                              disk, ui['global_force'],
                              in_d('tribo_system_name', ui))   
    return ball, disk, tribo_system
    """
    pass


def setup_pin_on_disk(ui):
    """Generate pin contact body (master; either ball or ring,
    depending on user input), then generate disk (slave)"""
    if ui['type_profile_cb1'] == Profiles.Circle.value:
        if ui['diameter_cb1'] == 2 * ui['profile_radius_cb1']:
            pin = Ball('pin', ui['e_cb1'], ui['ny_cb1'], ui['diameter_cb1'],
                       rms_rough=in_d('rms_roughness_cb1', ui))
            pin.make_profile(ui['res_x'], ui['res_y'],
                             ui['global_force'] / ui['number_pins'])
        else:
            pin = Ring('pin', ui['e_cb1'], ui['ny_cb1'], ui['diameter_cb1'],
                       ui['length_cb1'], ui['type_profile_cb1'],
                       rms_rough=in_d('rms_roughness_cb1', ui),
                       circle_radius=in_d('profile_radius_cb1', ui),
                       path_profile=in_d('path_profile_cb1', ui))
            pin.make_profile(ui['res_x'], ui['res_y'],
                             ui['global_force'] / ui['number_pins'], 1)
    else:
        pin = Ring('pin', ui['e_cb1'], ui['ny_cb1'], ui['diameter_cb1'],
                   ui['length_cb1'], ui['type_profile_cb1'],
                   rms_rough=in_d('rms_roughness_cb1', ui),
                   circle_radius=in_d('profile_radius_cb1', ui),
                   path_profile=in_d('path_profile_cb1', ui))
        pin.make_profile(ui['res_x'], ui['res_y'],
                         ui['global_force'] / ui['number_pins'], 1)
    disk = Disk('disk', ui['e_cb2'], ui['ny_cb2'],
                rms_rough=in_d('rms_roughness_cb2', ui))
    disk.make_slave_to(pin)
    tribo_system = PinOnDisk(ui['number_pins'], ui['sliding_diameter'], pin,
                             disk, ui['global_force'],
                             in_d('tribo_system_name', ui))
    return pin, disk, tribo_system


def setup_four_ball(ui):
    """Generate rotating ball (master) and stationary ball (slave) contact
    bodies"""
    rot_ball = Ball('rotating-ball', ui['e_cb1'], ui['ny_cb1'],
                    ui['diameter_cb1'],
                    rms_rough=in_d('rms_roughness_cb1', ui))
    rot_ball.make_profile(ui['res_x'], ui['res_y'],
                          ui['global_force'] / math.sqrt(6))
    stat_ball = Ball('stationary-ball', ui['e_cb2'], ui['ny_cb2'],
                     ui['diameter_cb2'],
                     rms_rough=in_d('rms_roughness_cb2', ui))
    stat_ball.make_slave_to(rot_ball)
    tribo_system = FourBall(rot_ball, stat_ball, ui['global_force'],
                            in_d('tribo_system_name', ui))
    return rot_ball, stat_ball, tribo_system


def setup_ball_on_three_plates(ui):
    """Generate ball (master) and flat (slave) contact bodies"""
    ball = Ball('ball', ui['e_cb1'], ui['ny_cb1'], ui['diameter_cb1'],
                rms_rough=in_d('rms_roughness_cb1', ui))
    ball.make_profile(ui['res_x'], ui['res_y'],
                      ui['global_force'] * math.sqrt(2) / 3)
    plate = Flat('plate', ui['e_cb2'], ui['ny_cb2'],
                 rms_rough=in_d('rms_roughness_cb2', ui))
    plate.make_slave_to(ball)
    tribo_system = BallOnThreePlates(ball, plate, ui['global_force'],
                                     in_d('plate_angle', ui, math.pi / 2),
                                     in_d('tribo_system_name', ui))
    return ball, plate, tribo_system


def setup_ring_on_ring(ui):
    """"Generate central ring contact body (sun) as well as planetary ring(s),
    then make planets slave to sun"""
    sun = Ring('sun', ui['e_cb1'], ui['ny_cb1'], ui['diameter_cb1'],
               ui['length_cb1'], ui['type_profile_cb1'],
               rms_rough=in_d('rms_roughness_cb1', ui),
               circle_radius=in_d('profile_radius_cb1', ui),
               path_profile=in_d('path_profile_cb1', ui))
    sun.make_profile(ui['res_x'], ui['res_y'],
                     ui['global_force'] / ui['number_planets'], 1)
    planet = Ring('planet', ui['e_cb2'], ui['ny_cb2'], ui['diameter_cb2'],
                  ui['length_cb2'], ui['type_profile_cb2'],
                  rms_rough=in_d('rms_roughness_cb2', ui),
                  circle_radius=in_d('profile_radius_cb2', ui),
                  path_profile=in_d('path_profile_cb2', ui))
    planet.make_slave_to(sun)
    tribo_system = RingOnRing(ui['number_planets'], sun, planet,
                              ui['global_force'], in_d('tribo_system_name', ui))
    return sun, planet, tribo_system



# -*- coding: utf-8 -*-

"""

This module contains functions related to Dowson-Hamrock film thickness
calculations.

"""


def __dowson_hamrock_parameters(r_eff, param_g, param_u, param_w):
    """

     Calculate the EHD-parameter for Dowson-Hamrock film thickness calculations.
     The name EHD-parameter is used within the tribology package to refer to
     what is calculated below; this parameter is not officially defined by
     Dowson-Hamrock.

    Parameters
    ----------
    r_eff: scalar
        The effective radius of the contact problem.
    param_g: scalar
        The elasticity parameter of the contact problem.
    param_u: scalar
        The velocity parameter of the contact problem.
    param_w: scalar
        The load parameter of the contact problem.

    Returns
    -------
    param_ehd: scalar
        The EHD parameter of the contact problem.

    """
    param_ehd = r_eff * param_g ** 0.53 * param_u ** 0.67 * param_w ** -0.067
    return param_ehd


def edowham(alpha_p, e_eff):
    """

    Calculate the elasticity parameter of the contact problem according to
    Dowson-Hamrock.

    Parameters
    ----------
    alpha_p: ndarray, scalar
        The pressure-viscosity coefficient of the lubricant.
    e_eff: ndarray, scalar
        The effective modulus of the contact problem.

    Returns
    -------
    param_elasticity: ndarray, scalar
        The elasticity parameter of the contact problem.

    """
    param_elasticity = alpha_p * e_eff
    return param_elasticity


def vdowham(eta, vel_entrain, e_eff, r_eff):
    """

    Calculate the velocity parameter of the contact problem according to
    Dowson-Hamrock.

    Parameters
    ----------
    eta: ndarray, scalar
        The dynamic viscosity of the lubricant.
    vel_entrain: ndarray, scalar
        The entrainment velocity of the contact problem.
    e_eff: ndarray, scalar
        The effective modulus of the contact problem.
    r_eff: ndarray, scalar
        The effective radius of the contact problem.

    Returns
    -------
    param_velocity: ndarray, scalar
        The velocity parameter of the contact problem.

    """
    param_velocity = eta * vel_entrain / (e_eff * r_eff)
    return param_velocity


def dowhamline(vel_entrain, force, alpha_p, e_eff, r_eff, eta, l_eff):
    """

    Calculate the mean film thickness in a line contact according to
    Dowson-Hamrock.

    Parameters
    ----------
    vel_entrain: ndarray, scalar
        The entrainment velocity of the contact problem.
    force: ndarray, scalar
        The normal force in the contact.
    alpha_p: ndarray, scalar
        The pressure-viscosity coefficient of the lubricant.
    e_eff: ndarray, scalar
        The effective modulus of the contact problem.
    r_eff: ndarray, scalar
        The effective radius of the contact problem.
    eta: ndarray, scalar
        The dynamic viscosity of the lubricant.
    l_eff: ndarray, scalar
        The effective length of the contact.

    Returns
    -------
    h_0: ndarray, scalar
        The central lubricating film thickness in the contact.

    """
    param_g = edowham(alpha_p, e_eff)
    param_u = vdowham(eta, vel_entrain, e_eff, r_eff)
    param_w = force / (l_eff * r_eff * e_eff)
    h_0 = 2.69 * __dowson_hamrock_parameters(r_eff, param_g, param_u, param_w)
    return h_0


def dowhampoint(vel_entrain, force, alpha_p, e_eff, r_eff, eta):
    """

    Calculate the mean film thickness in a point contact according to
    Dowson-Hamrock.

    Parameters
    ----------
    vel_entrain: ndarray, scalar
        The entrainment velocity of the contact problem.
    force: ndarray, scalar
        The normal force in the contact.
    alpha_p: ndarray, scalar
        The pressure-viscosity coefficient of the lubricant.
    e_eff: ndarray, scalar
        The effective modulus of the contact problem.
    r_eff: ndarray, scalar
        The effective radius of the contact problem.
    eta: ndarray, scalar
        The dynamic viscosity of the lubricant.

    Returns
    -------
    h_0: ndarray, scalar
        The central lubricating film thickness in the contact.

    """
    param_g = edowham(alpha_p, e_eff)
    param_u = vdowham(eta, vel_entrain, e_eff, r_eff)
    param_w = force / (r_eff ** 2 * e_eff)
    h_0 = 1.9 * __dowson_hamrock_parameters(r_eff, param_g, param_u, param_w)
    return h_0



# -*- coding: utf-8 -*-

"""

The functions in this module are related to tribology in general.

"""

from math import sqrt, acos, cos, sin, pi, floor

import numpy as np


def profrolleriso(x_axis, diam, length):
    """

    Generate a roller profile according to DIN 26281 for rollers with
    :code:`length` :math:`\\leq` :code:`2.5` :math:`\\cdot` :code:`diam`.

    Parameters
    ----------
    x_axis: ndarray
        The x-axis of the roller. The origin of the axis must be at the center
        of the roller. To obtain sensible results, the axis limits must be
        within :code:`+/- length / 2`.
    diam: scalar
        The diameter of the roller.
    length: scalar
        The length of the roller.

    Returns
    -------
    x_profile: ndarray
        The profile heights of the roller along :code:`x_axis`.

    """
    len_ax = abs(max(x_axis) - min(x_axis))
    if len_ax == length:
        ax_x = x_axis[1:-1]
    else:
        ax_x = x_axis

    prof = 0.00035 * diam * np.log(1 / (1 - np.power(((2 * ax_x) / length), 2)))
    prof += np.min(prof)

    if len_ax == length:
        diam_arr = np.asarray([diam])
        x_profile = np.concatenate((diam_arr, np.concatenate((prof, diam_arr))))
    else:
        x_profile = prof

    return x_profile


def profball(x_axis, r_ball):
    """

    Create a 1D ball profile along a central axis.

    Parameters
    ----------
    x_axis: ndarray
        The coordinate points for which to calculate profile heights. Minimum
        and maximum value must be within +/- **r_ball**.
    r_ball: scalar
        The radius of the ball.

    Returns
    -------
    prof_heights: ndarray
        The profile heights along the **x_axis**.

    """
    prof = np.array(abs(r_ball) - np.sqrt(r_ball ** 2 - np.power(x_axis, 2)))
    prof_heights = prof * np.sign(r_ball)
    return prof_heights


def profrevolve(prof_1d, y_axis, y_diam):
    """

    Create a 2D surface (profile heights) by revolving a 1D profile around a
    central axis.

    Parameters
    ----------
    prof_1d: ndarray
        The profile heights along the central (x) axis of the surface.
    y_axis: ndarray
        The y-axis of the surface.
    y_diam: scalar
        The maximum diameter of the surface along its y-axis.

    Returns
    -------
    prof_2d: ndarray
        The surface heights of the profile. The array size is
        :code:`len(prof_2d)` :math:`\\times` :code:`len(y_axis)`
    y_profile: ndarray
        The profile heights along the y-direction of the surface at the central
        element of the x-axis.

    """
    len_x = len(prof_1d)
    len_y = len(y_axis)
    prof_2d = np.zeros((len_x, len_y))
    sign_diam = np.sign(y_diam)
    for i_x in range(len_x):
        for i_y in range(len_y):
            bracket = pow((y_diam / 2 - sign_diam * prof_1d[i_x]), 2)\
                      - pow(y_axis[i_y], 2)
            if bracket <= 0:
                prof_2d[i_x, i_y] = abs(y_diam) / 2
            else:
                prof_2d[i_x, i_y] = abs(y_diam) / 2 - sign_diam * sqrt(bracket)
    if y_diam > 0:
        min_prof = prof_2d.min()
    else:
        min_prof = prof_2d[round(len_x / 2) - 1, round(len_x / 2) - 1]
    prof_2d = (prof_2d - min_prof)
    y_profile = prof_2d[:, floor(len_y / 2)]
    return prof_2d, y_profile


def vslide(vel_1, vel_2):
    """

    Calculate the sliding speed in a tribological contact based on contact body
    velocities.

    Parameters
    ----------
    vel_1: ndarray, scalar
        The contact velocity of body 1.
    vel_2: ndarray, scalar
        The contact velocity of body 2

    Returns
    -------
    vel_slide: ndarray, scalar
        The sliding velocity in the tribological contact.

    """
    vel_slide = vel_1 - vel_2
    return vel_slide


def vroll(vel_1, vel_2):
    """

    Calculate the rolling speed in a tribological contact based on contact body
    velocities.

    Parameters
    ----------
    vel_1: ndarray, scalar
        The contact velocity of body 1.
    vel_2: ndarray, scalar
        The contact velocity of body 2

    Returns
    -------
    vel_roll: ndarray, scalar
        The rolling velocity in the tribological contact.

    """
    vel_roll = (vel_1 + vel_2) / 2
    return vel_roll


def srr(vel_1, vel_2):
    """

    Calculate the slide-to-roll ratio (SRR) in a tribological contact based on
    contact body velocities.

    Parameters
    ----------
    vel_1: ndarray, scalar
        The contact velocity of body 1.
    vel_2: ndarray, scalar
        The contact velocity of body 2

    Returns
    -------
    srr: ndarray, scalar
        The slide-to-roll ratio in the tribological contact.

    """
    val_srr = vslide(vel_1, vel_2) / vroll(vel_1, vel_2)
    return val_srr


def radpersec2rpm(vel_rad_per_sec):
    """

    Convert velocity from rotations per minute (rpm) to radians per second
    (rad/s).


    Parameters
    ----------
    vel_rad_per_sec: ndarray, scalar
        The velcity in radians per second.

    Returns
    -------
    vel_rpm: ndarray, scalar
        The velocity in rotations per minute.

    """
    vel_rpm = 1 / rpm2radpersec(1 / vel_rad_per_sec)
    return vel_rpm


def rpm2radpersec(vel_rpm):
    """

     Convert velocity from radians per second (rad/s) to rotations per minute
     (rpm).


     Parameters
     ----------
     vel_rpm: ndarray, scalar
         The velcity in rotations per minute.

     Returns
     -------
     vel_rad_per_sec: ndarray, scalar
         The velocity in radians per second.

     """
    vel_rad_per_sec = vel_rpm / 60 * 2 * pi
    return vel_rad_per_sec


def rball3plates(r_ball, plate_angle=1.5708):
    """

    Calculate the sliding radius (lever arm) for a ball-on-3-plates test setup.

    Parameters
    ----------
    r_ball: ndarray, scalar
        The radius of the ball.
    plate_angle: ndarray, scalar, optional
        The plate angle in radians, measured with respect to the rotational
        axis of the ball. Default value corresponds to 45 degree.

    Returns
    -------
    r_slide: ndarray, scalar
        The sliding radius.

    """
    r_slide = r_ball * np.sin((pi - plate_angle) / 2)
    return r_slide


def fball3plates(ax_force, plate_angle=1.5708):
    """

    Calculate the normal force per contact in ball-on-3-plates setup.

    Parameters
    ----------
    ax_force: ndarray, scalar
        The force acting along the rotational axis of the ball (axial force).
    plate_angle: ndarray, scalar, optional
        The plate angle in radians, measured with respect to the rotational
        axis of the ball. Default value corresponds to 45 degree.

    Returns
    -------
    norm_force:
        The normal force acting in each ball-plate contact.

    """
    norm_force = ax_force / 3 / np.cos(plate_angle / 2)
    return norm_force


def gfourball(r_1, r_2):
    """

    Geometric parameters of 4-ball setup

    Parameters
    ----------
    r_1 : scalar
        The radius of the rotating ball.
    r_2 : scalar
        The radius of the (a) stationary ball.

    Returns
    -------
    sliding_radius : scalar
        The sliding radius (lever arm) on the rotating ball.
    contact_angle : scalar
        Contact angle between direction of ball normal force and vertical axis.

    """
    r_circum_circle = sqrt(3) / 3 * 2 * r_2
    contact_angle = acos(r_circum_circle / (r_1 + r_2))
    sliding_radius = r_circum_circle - r_2 * cos(contact_angle)
    return sliding_radius, contact_angle


def ffourball(r_1, r_2, ax_force):
    """

    Calculate the normal force per contact in a 4-ball test setup.

    Parameters
    ----------
    r_1:
        The radius of the rotating ball.
    r_2:
        The radius of the stationary balls.
    ax_force:
        The force acting along the rotational axis of the rotating ball.

    Returns
    -------
    norm_force: ndarray, scalar
        The normal force in a single ball-ball contact.

    """
    _, contact_angle = gfourball(r_1, r_2)
    norm_force = ax_force / sin(contact_angle) / 3
    return norm_force


def refix(val, p_in="", p_out=""):
    """

    Convert between different SI unit prefixes. Available options are:

    :code:`'T'`   Terra

    :code:`'G'` Giga

    :code:`'M'` Mega

    :code:`'k'` Kilo

    :code:`'m'` Milli

    :code:`'mu'` Micro

    :code:`'n'` Nano

    :code:`'p'` Pico

    Parameters
    ----------
    val: scalar
        The value for which to convert the unit prefix.
    p_in: string, any of the above, optional
        The current prefix of :code:`val`. If :code:`p_in` is undefined,
        :code:`val` has no SI unit prefix.
    p_out: string, any of the above, optional
        The prefix of :code:`val_refix` after the conversion.  If :code:`p_in`
        is undefined, :code:`val_refix` has no SI unit prefix.

    Returns
    -------
    val_refix: scalar
        The value in units of prefix :code:`p_out`.

    """
    prefix = {'p': 10 ** -12,
              'n': 10 ** -9,
              'mu': 10 ** -6,
              'm': 10 ** -3,
              '': 10 ** -0,
              'k': 10 ** 3,
              'M': 10 ** 6,
              'G': 10 ** 9,
              'T': 10 ** 12}
    val_refix = val * prefix[p_in] / prefix[p_out]
    return val_refix


def abbottfirestone(trace, num_bins=100):
    """

    Calculate the Abbott-Firestone curve for a 1D profile trace.

    Parameters
    ----------
    trace: ndarray
        The profile heights of the 1D trace.
    num_bins: positive int, optional
        The number of bins for the calculation of the profile height
        probability distribution.

    Returns
    -------
    bins: ndarray
        The bins used for the calculation of the profile height
        probability distribution (= x-axis data for Abbott-Firestone plot).
    prob_dist: ndarray
        The probability distribution of the profile (= y-axis data for Abbott-
        Firestone plot).

    """
    bins = np.linspace(np.amax(trace), np.amin(trace), num_bins)
    prob_dist = []
    for each_bin in bins:
        prob_dist.append(
            len(np.where(trace >= each_bin)[0]) / len(trace) * num_bins)
    return bins, prob_dist


if __name__ == "__main__":
    pass



"""

The below functions are not automatically imported when you import the
tribology package. They need to be imported on a sub-module level:

:code:`from tribology.p3can.p3can import p3can, generate_input_file`

**Note: The P3CAN project is currently not maintained!**

The P3CAN project aimed to develop a Python 3 implementation of an open source
multi-purpose (tribology) contact analyzer, both for research and higher level
educational purposes. However, due to its monolithic structure it proved hard
to maintain, and it can be hard for new users to understand the code in its
entirety.

**The P3CAN project precedes the** :code:`tribology` **package and many of its
functions and functionalities have since been integrated into the package. This
work will be continued in the future.**

Simply put, P3CAN can be used to calculate parameters such
as contact pressure, load distribution and friction energy for various
tribological systems.

P3can currently provides implementations of the following tribological systems:
- radial roller bearings
- axial thrust roller bearings
- pin-on-disk
- 4-ball
- ball-on-3-plates
- ring-on-ring

You can run P3CAN by calling it with a program specific input file. Templates
for user input files can be generated using the `generate_input_file`
function below.

For more information on the P3CAN project, please refer to its GitHub page:
https://github.com/moritzploss/p3can

"""

import os
import shutil

from BluPrintSim import Sim
from Constants import SimType
from generate_latex_output import generate_latex_output
from influ_matrix_management import make_infl_mat_db, manage_influ_mat_cache
from setup_tribosystem import setup_cyl_rol_bearing, \
    setup_deep_gro_ball_bearing, setup_cyl_rol_thrust_bearing, \
    setup_ball_on_disk, setup_pin_on_disk, setup_four_ball, \
    setup_ball_on_three_plates, setup_ring_on_ring
from system_functions import exit_program, print_it, in_d


def p3can(in_file='USER_INPUT.py', out_dir=None):
    """

    Start a P3CAN simulation run.

    Parameters
    ----------
    in_file: str, optional
        The file path of the P3CAN input file. The default value is
        `USER_INPUT.py`. You can generate `in_file` templates for different
        tribological systems using the `generate_input_file` function.
    out_dir: str
        The directory in which output files will be stored. A folder named
        `results` will be created in `out_dir`. The `results` folder then
        contains a simulation-specific folder containing all outputs. Default
        is the current working directory.

    Returns
    -------
    sim.res_dir: str
        The path to the simulation-specific results folder. The path is:

        `out_dir`/results/<unique simulation name>

        The unique simulation name is automatically generated by P3CAN.

    """

    if not out_dir:
        out_dir = os.getcwd()

    # set-up a simulation
    sim = Sim(in_file, out_dir)
    sim.mk_uniq_parameter_id()
    sim.infl_db_file_hand = make_infl_mat_db(os.path.dirname(sim.res_dir))
    tribo_system = None

    # generate a tribosystem based on user input
    print_it('initialising contact bodies')
    if sim.simulation_type == SimType.cyl_rol_bearing.value:
        roller, ring1, ring2, tribo_system = setup_cyl_rol_bearing(sim.ui)
    elif sim.simulation_type == SimType.deep_gro_ball_bearing.value:
        setup_deep_gro_ball_bearing(sim.ui)
    elif sim.simulation_type == SimType.cyl_rol_thrust_bearing.value:
        roller, ring1, ring2, tribo_system = setup_cyl_rol_thrust_bearing(
            sim.ui)
    elif sim.simulation_type == SimType.ball_on_disk.value:
        ball, disk, tribo_system = setup_ball_on_disk(sim.ui)
    elif sim.simulation_type == SimType.pin_on_disk.value:
        pin, disk, tribo_system = setup_pin_on_disk(sim.ui)
    elif sim.simulation_type == SimType.four_ball.value:
        rot_ball, stat_ball, tribo_system = setup_four_ball(sim.ui)
    elif sim.simulation_type == SimType.ball_on_three_plates.value:
        ball, plate, tribo_system = setup_ball_on_three_plates(sim.ui)
    elif sim.simulation_type == SimType.ring_on_ring.value:
        sun, planet, tribo_system = setup_ring_on_ring(sim.ui)
    else:
        exit_program("'simulation_type' option '{}'not defined".format(
            sim.simulation_type))

    # simulate tribosystem
    tribo_system.calc_load_distribution(sim.ui, sim.res_dir)
    tribo_system.calc_contact_pressure(sim.ui, sim.res_dir)
    tribo_system.calc_kinematics(in_d('rot_velocity1', sim.ui, 0),
                                 in_d('rot_velocity2', sim.ui, 0),
                                 sim.ui, sim.res_dir)
    tribo_system.calc_pv(sim.ui, sim.res_dir)
    tribo_system.calc_e_akin(sim.ui, sim.res_dir)

    # generate output plots and pdf report
    if sim.auto_plot is True:
        tribo_system.plot_it(sim.ui, sim.res_dir)
    if sim.auto_report is True:
        generate_latex_output(sim, tribo_system, sim.ui, sim.res_dir)

    # manage the influence matrix cache
    manage_influ_mat_cache(sim.res_dir)

    # finish simulation
    sim.finished()
    return sim.res_dir


def generate_input_file(temp_type, out_file):
    """

    Generate a P3CAN input file based on a user input template.

    Parameters
    ----------
    temp_type: int
        Specifies which type of user input template will be generated:

        0:  complete set of available parameters

        1:  template for single row cylindrical roller bearings

        3:  template for cylindrical roller thrust bearings

        4:  template for ball-on-disk setup

        5:  template for pin-on-disk setup

        6:  template for 4-ball setup

        7:  template for ball-on-3-plates setup

        8:  template for ring-on-ring setup

    out_file: str
        The complete file path of the output file.

    Returns
    -------
    out_file: str
        The complete file path of the output file.

    """

    file_path = os.path.realpath(__file__)
    dir_path = os.sep.join(file_path.split(os.sep)[:-1])

    if temp_type == 0:
        template = 'Template00_CompleteParameters.py'
    elif temp_type == 1:
        template = 'Template01_SingleRowCylindricalRollerBearing.py'
    elif temp_type == 3:
        template = 'Template03_CylindricalRollerThustBearing.py'
    elif temp_type == 4:
        template = 'Template04_BallOnDisk.py'
    elif temp_type == 5:
        template = 'Template05_PinOnDisk.py'
    elif temp_type == 6:
        template = 'Template06_4Ball.py'
    elif temp_type == 7:
        template = 'Template07_BallOn3Plates.py'
    elif temp_type == 8:
        template = 'Template08_RingOnRing.py'
    else:
        raise ValueError("temp_type value '{}' undefined".format(temp_type))

    shutil.copy(os.sep.join([dir_path, 'UserInputTemplates', template]),
                out_file)
    return out_file


if __name__ == "__main__":
    pass



import sys

from Constants import PrintOpts
from system_functions import print_it, exit_program


def check_lapack():
    """Check if BLAS/LAPACK distribution is installed"""
    import numpy as np
    if not np.__config__.lapack_opt_info or not np.__config__.blas_opt_info:
        print_it('could not find BLAS/LAPACK install\n'
                 'install BLAS/LAPACK to improve p3can performance',
                 PrintOpts.lvl1.value)
        print_it('did not check for ATLAS', PrintOpts.lvl1.value)
    else:
        print_it('found BLAS/LAPACK install', PrintOpts.lvl1.value)
        print_it('did not check for ATLAS', PrintOpts.lvl1.value)


def check_python_environment():
    """Check if all required python environments are installed before actual
    calculation is started"""
    print_it('python: {}'.format(sys.version.replace('\n', ' ')),
             PrintOpts.lvl1.value)
    try:
        import matplotlib as mpl
        print_it('matplotlib: {}'.format(mpl.__version__),
                 PrintOpts.lvl1.value)
    except ImportError:
        print_import_error('matplotlib', PrintOpts.lvl1.value)
        exit_program('module not found')
    try:
        import numexpr
        print_it('numexpr: {}'.format(numexpr.__version__),
                 PrintOpts.lvl1.value)
    except ImportError:
        print_import_error('numexpr', PrintOpts.lvl1.value)
        exit_program('module not found')
    try:
        import numpy
        print_it('numpy: {}'.format(numpy.__version__),
                 PrintOpts.lvl1.value)
    except ImportError:
        print_import_error('numpy', PrintOpts.lvl1.value)
        exit_program('module not found')
    try:
        import scipy
        print_it('scipy: {}'.format(scipy.__version__),
                 PrintOpts.lvl1.value)
    except ImportError:
        print_import_error('scipy', PrintOpts.lvl1.value)
        exit_program('module not found')
    try:
        import jinja2
        print_it('jinja2: {}'.format(jinja2.__version__),
                 PrintOpts.lvl1.value)
    except ImportError:
        print_import_error('jinja2', PrintOpts.lvl1.value)
        print_it('this will cause an error if auto_report is True')


def print_import_error(module_name, level):
    """Print import error message"""
    print_it("could not find (import) module '{}'".format(module_name), level)
    print_it("the module is missing or unknown to the python interpreter",
             level)




def secant_method(x, fx):
    """Applies secant method to find root x of function f(x). If not enough
    (x, f(x)) value pairs are known to apply secant method, a new x value is
    guessed by slightly changing the initial x value"""
    if fx[-1] != 0:
        if len(x) > 1 and fx[-1] != 0 and abs(fx[-1]) != abs(fx[-2]):
            x0 = x[-2]
            x1 = x[-1]
            fx0 = fx[-2]
            fx1 = fx[-1]
            m = (fx1 - fx0) / (x1 - x0)
            return x1 + (-fx1 / m)
        else:
            return x[0] * 0.9 + 0.0001
    else:
        return x[-1]



import math

import numpy as np

from Constants import SubDir
from system_functions import make_directory
import matplotlib.pyplot as plt


def generate_pub_figures(sys, sim):
    """Generate summary figure for Cylindrical Roller Bearing Tribosystems"""
    plot_axis_pol = np.linspace(-math.pi, math.pi - 2 * math.pi / sys.res_pol,
                                int(sys.res_pol))
    pv_max_ring1 = np.concatenate(
        (sys.pv_ring1[round(sys.res_pol / 2) - 1:-1, :],
         sys.pv_ring2[0:round(sys.res_pol / 2), :]), axis=0)
    eakin_data = np.concatenate(
        (sys.e_akin_ring2.transpose()[round(sys.res_pol / 2) - 1:-1, :],
         sys.e_akin_ring2.transpose()[0:round(sys.res_pol / 2), :]), axis=0)
    pressure_data = np.concatenate(
        (sys.max_pressure_ring1[round(sys.res_pol / 2) - 1:-1, :],
         sys.max_pressure_ring1[0:round(sys.res_pol / 2), :]), axis=0)

    plt.figure()
    ax = plt.subplot(111, polar=True, facecolor='White')
    ax.grid(True)
    ax.plot(plot_axis_pol,
            pv_max_ring1[:, round(sys.roller.res_x / 2) - 1] / np.amax(
                pv_max_ring1), 'k-', label='pv')
    ax.plot(plot_axis_pol,
            eakin_data[:, round(sys.roller.res_x / 2) - 1] / np.amax(
                eakin_data), 'k--', label='eakin')
    ax.plot(plot_axis_pol,
            pressure_data[:, round(sys.roller.res_x / 2) - 1] / np.amax(
                pressure_data), 'k-.',
            label='max pressure')
    ax.plot(sys.pol_ax, sys.normal_forces / np.amax(sys.normal_forces), 'k:',
            label='normal force')
    ax.set_theta_zero_locatioaxisbgn("S")
    ax.set_rlabel_position(112.5)
    label_position = ax.get_rlabel_position()
    r_max = 1.1
    r_label = 'normalised value'
    ax.text(math.radians(label_position + 15), r_max / 2, r_label,
            rotation=22.5, ha='center', va='center')
    ax.set_rmax(r_max)
    make_directory(sim.results_folder, SubDir.pub_figs.value)
    file_handle = sim.results_folder + '/' + SubDir.pub_figs.value + '/' + \
                  'normalised-inner-ring' + '.png'
    plt.savefig(file_handle, bbox_inches='tight', dpi=300)
    plt.close()



import math
import os
import re

import numpy as np

from BluPrintTriboSys import TriboSys
from Constants import PltOpts, SubDir, TexTempl, UnitTex, Unit, PrintOpts, \
    PreSol
from cartesian_plot_functions import plt_profile, plt_contact, plt_3d, \
    plt_2d_scatt_line, \
    plt_energy_ring_on_ring, plt_profile_approx
from generate_latex_output import get_calc_specific_latex_template
from hertz_equations import hertz_displ
from influ_matrix_management import load_influ_mat, cache_influ_mat
from solve_half_space import solve_half_space, pre_solve_half_space
from system_functions import print_it, to_preci, exit_program, save_to_matlab


class RingOnRing(TriboSys):
    """Global tribosystem data"""

    def __init__(self, num_planets, sun, planet, global_force,
                 setup_name='ring-on-ring'):
        super().__init__(setup_name, global_force, None)

        self.num_planets = num_planets
        self.sun = sun
        self.planet = planet
        self.sun.norm_forces = None
        self.init_force = None

        self.planet_slip = None
        self.rot_velocity = None
        self.rot_velocity2 = None
        self.sun_rot_vel = None
        self.planet_rot_vel = None
        self.sliding_vel = None
        self.sun.press = None
        self.sun.max_press = None

        self.sun.rot_vel = None
        self.planet.rot_vel = None
        self.sun.omega = None
        self.planet.omega = None
        self.sun.vel = None
        self.planet.vel = None
        self.slip = None
        self.rel_vel = None
        self.pv = None

        self.influ_mat_db_1 = None
        self.press_zone_len = None

    def calc_load_distribution(self, ui=None, res_dir=None):
        """Calculate load distribution"""
        print_it("calculating load distribution")
        self.sun.norm_forces = np.multiply(np.ones(self.num_planets),
                                           self.global_force / self.num_planets)
        self.init_force = self.global_force

    def get_grid_size(self, ui, res_dir):
        """Determine grid size by running (quick) simulation with simplified
        contact bodies"""
        print_it('determining grid size', PrintOpts.lvl1.value)
        self.sun.simple_clone()
        self.sun.clone.make_profile(PreSol.res_x.value, PreSol.res_y.value,
                                    self.init_force)
        self.planet.simple_clone()
        self.planet.clone.make_slave_to(self.sun.clone)

        init_displ = hertz_displ(self.sun.clone.e, self.planet.e,
                                 self.sun.clone.ny, self.planet.ny,
                                 self.sun.clone.r_hertz_x,
                                 self.sun.clone.r_hertz_y,
                                 self.planet.clone.r_hertz_x,
                                 self.planet.clone.r_hertz_y,
                                 self.sun.norm_forces[0])
        too_many_els_in_y = 1
        too_many_els_in_x = 1
        contact_width_y = 0.05
        contact_width_x = 0.05
        while too_many_els_in_y != 0 or \
              too_many_els_in_x != 0:
            self.sun.clone.make_profile(self.sun.clone.res_x,
                                        self.sun.clone.res_y, self.init_force,
                                        contact_width=contact_width_y,
                                        contact_length=contact_width_x)
            self.planet.clone.make_slave_to(self.sun.clone)

            pressure, init_displ = \
                pre_solve_half_space(self.sun.clone.profile,
                                     self.planet.clone.profile,
                                     self.sun.clone.x_axis,
                                     self.sun.clone.y_axis,
                                     self.sun.clone.res_x, self.sun.clone.res_y,
                                     self.sun.clone.delta_x,
                                     self.sun.clone.delta_y, self.sun.clone.e,
                                     self.planet.clone.e, self.sun.clone.ny,
                                     self.planet.clone.ny,
                                     self.sun.norm_forces[0],
                                     init_displ=init_displ, print_prog=False)

            pressure_els_y = sum(
                pressure[math.floor(self.sun.clone.res_y / 2), :] > 0)
            too_many_els_in_y = self.sun.clone.res_y - pressure_els_y - 2
            if too_many_els_in_y:
                contact_width_y += -np.sign(
                    too_many_els_in_y) * contact_width_y / 25

            pressure_els_x = sum(
                pressure[:, math.floor(self.sun.clone.res_x / 2)] > 0)
            too_many_els_in_x = self.sun.clone.res_x - pressure_els_x - 2
            if too_many_els_in_x:
                contact_width_x += -np.sign(
                    too_many_els_in_x) * contact_width_x / 25

        self.sun.make_profile(self.sun.res_x, self.sun.res_y, self.init_force,
                              contact_width=contact_width_y,
                              contact_length=contact_width_x)
        self.planet.make_slave_to(self.sun)
        return init_displ

    def calc_contact_pressure(self, ui=None, res_dir=None):
        """Calculate contact pressure distribution between sun and planet
        ring(s)"""
        print_it('calculating 1 pressure distribution')
        init_displ = self.get_grid_size(ui, res_dir)
        [self.influ_mat_db_1] = load_influ_mat(ui, res_dir, 1)
        print_it('solving first half space', PrintOpts.lvl1.value)
        self.sun.press, self.influ_mat_db_1 = \
            solve_half_space(self.sun.profile, self.planet.profile,
                             self.sun.x_axis, self.sun.y_axis, self.sun.res_x,
                             self.sun.res_y, self.sun.delta_x, self.sun.delta_y,
                             self.sun.e, self.planet.e,
                             self.sun.ny, self.planet.ny,
                             self.sun.norm_forces[0], res_dir,
                             init_displ=init_displ,
                             influ_mat_db=self.influ_mat_db_1)

        cache_influ_mat(ui, [self.influ_mat_db_1], res_dir)
        self.sun.max_press = np.amax(self.sun.press, axis=1)

        dat_dict = dict(x_axis=self.sun.x_axis,
                        y_axis=self.sun.y_axis,
                        contact_pressure=self.sun.press)
        save_to_matlab(dat_dict, res_dir, 'pressure_field')

    def calc_kinematics(self, rot_vel1, rot_vel2, ui=None, res_dir=None):
        """Calculate tribosystem kinematics based on rotational velocities of
        sun and planet(s)"""
        print_it("calculating kinematics")
        self.sun.rot_vel = rot_vel1
        self.planet.rot_vel = rot_vel2
        self.sun.omega = self.sun.rot_vel / 60
        self.planet.omega = self.planet.rot_vel / 60
        self.sun.vel = self.sun.diameter * math.pi * self.sun.omega
        self.planet.vel = self.planet.diameter * math.pi * self.planet.omega
        self.slip = (self.sun.vel - self.planet.vel) / self.sun.vel
        self.rel_vel = np.ones(self.sun.res_x) * (
        self.sun.vel - self.planet.vel)

        self.sun.footpr_vel = \
            2 * math.pi * self.sun.diameter / 2 * self.sun.omega
        self.planet.footpr_vel = \
            2 * math.pi * self.planet.diameter / 2 * self.planet.omega
        try:
            self.sun.overroll_t_incr = self.sun.delta_y / self.sun.footpr_vel
            self.planet.overroll_t_incr = \
                self.planet.delta_y / self.planet.footpr_vel
        except ZeroDivisionError:
            exit_program(
                'rotational velocities of sun and planet must not be 0')

        self.press_zone_len = (self.sun.press > 0).sum(1) * self.sun.delta_y
        self.sun.overroll_t = np.divide(self.press_zone_len,
                                        self.sun.footpr_vel)
        self.planet.overroll_t = np.divide(self.press_zone_len,
                                           self.planet.footpr_vel)
        self.sun.no_overroll_t = np.divide(
            (2 * math.pi * (self.sun.diameter / 2) - self.num_planets *
             self.press_zone_len) / self.num_planets, self.sun.footpr_vel)
        self.planet.no_overroll_t = np.divide(
            (2 * math.pi * (self.planet.diameter / 2) - self.press_zone_len),
            self.planet.footpr_vel)

    def calc_pv(self, ui=None, res_dir=None):
        """Calculate product of local maximum pressure and local maximum
        relative velocity"""
        print_it("calculating pv_rel")
        self.pv = np.multiply(abs(self.rel_vel), self.sun.max_press) / 1000
        dat_dict = dict(x_axis=self.sun.x_axis,
                        pv_rel=self.pv)
        save_to_matlab(dat_dict, res_dir, 'pv-rel')

    def calc_e_akin(self, ui=None, res_dir=None):
        """"Calculate the kinetic friction energy accumulation in W per m^2"""
        print_it("calculating e_a,kin")
        pv_local = np.multiply(self.sun.press.sum(1), self.rel_vel)
        self.sun.e_akin = np.absolute(
            np.divide(np.multiply(pv_local, self.sun.overroll_t_incr),
                      self.sun.no_overroll_t)) / 1000
        self.planet.e_akin = np.absolute(
            np.divide(np.multiply(pv_local, self.planet.overroll_t_incr),
                      self.planet.no_overroll_t)) / 1000

    def plot_it(self, ui=None, res_dir=None):
        """Orchestrate output plot generation"""
        print_it("plotting results")
        plt_profile(self.sun, PltOpts.DD.value, res_dir, SubDir.profiles.value)
        plt_profile(self.sun, PltOpts.DDD.value, res_dir, SubDir.profiles.value)
        plt_profile(self.planet, PltOpts.DD.value, res_dir,
                    SubDir.profiles.value)
        plt_profile(self.planet, PltOpts.DDD.value, res_dir,
                    SubDir.profiles.value)
        plt_profile_approx(res_dir, SubDir.profiles.value)
        plt_contact(self.sun, self.planet, PltOpts.DD.value, res_dir,
                    SubDir.contacts.value)
        plt_contact(self.sun, self.planet, PltOpts.DDD.value, res_dir,
                    SubDir.contacts.value)

        plt_3d(self.sun.x_axis, self.sun.y_axis, self.sun.press,
               self.sun.x_label, self.sun.y_label, 'pressure in MPa',
               'contact_pressure_sun', res_dir, SubDir.pressures.value,
               'contact_pressure_sun')
        plt_2d_scatt_line(self.sun.x_axis, self.pv, self.sun.x_axis, self.pv,
                          self.sun.x_label,
                          'pv_rel in {}'.format(Unit.pvrel.value), 'pv_rel',
                          res_dir, SubDir.energy.value, 'pv_rel')
        plt_2d_scatt_line(self.sun.x_axis, self.sun.e_akin, self.sun.x_axis,
                          self.sun.e_akin, self.sun.x_label,
                          'e_akin in {}'.format(Unit.eakin.value), 'e_akin',
                          res_dir, SubDir.energy.value, 'sun.e_akin')
        plt_2d_scatt_line(self.planet.x_axis, self.planet.e_akin,
                          self.planet.x_axis, self.planet.e_akin,
                          self.planet.x_label,
                          'e_akin in {}'.format(Unit.eakin.value), 'e_akin',
                          res_dir,
                          SubDir.energy.value, 'planet.e_akin')
        plt_energy_ring_on_ring(self, res_dir, SubDir.energy.value,
                                'e-akin-vs-pv-rel')

    def generate_latex_output(self, calc_spec_tex_file_handle, sim, ui=None,
                              res_dir=None):
        """Generate calculation-specific part of the LaTeX output file"""
        average_pressure = np.mean(self.sun.press[self.sun.press > 0])
        numeric_output_data = [
            ('pressure, max.', to_preci(np.amax(self.sun.press), 4),
             UnitTex.pressure.value, 'unverified'),
            ('pressure, av.', to_preci(average_pressure, 4),
             UnitTex.pressure.value, 'unverified'),
            ('e_a,kin sun, max.', to_preci(np.amax(self.sun.e_akin), 4),
             UnitTex.eakin.value, 'unverified'),
            ('e_a,kin planet, max.', to_preci(np.amax(self.planet.e_akin), 4),
             UnitTex.eakin.value, 'unverified'),
            ('pv_rel, max.', to_preci(np.amax(self.pv), 4),
             UnitTex.pvrel.value, 'unverified'),
            ('contact area', to_preci(self.sun.get_area(self.sun.press), 4),
             UnitTex.area.value, 'unverified')]
        table_calc_summary = []
        for key, value, unit, status in sorted(numeric_output_data):
            table_calc_summary.append(
                (re.sub('_', '\_', key), value, unit, status))

        latex_variables = {'table_calc_summary': table_calc_summary,
                           'contact_plot1': '{}{}contact1.png'.format(
                               SubDir.tex_figs_rel_to_tex_file.value, '/'),
                           'pressure_plot1': '{}{}pressure1.png'.format(
                               SubDir.tex_figs_rel_to_tex_file.value, '/'),
                           'energy_plot1': '{}{}energy1.png'.format(
                               SubDir.tex_figs_rel_to_tex_file.value, '/')}
        template_calc_specific = get_calc_specific_latex_template(
            TexTempl.RingOnRing.value, sim)
        with open(calc_spec_tex_file_handle, 'w') as f:
            f.write(template_calc_specific.render(latex_variables))

    def generate_latex_figures(self, ui=None, res_dir=None):
        """Generate calculation-specific figures for LaTeX report"""
        plt_contact(self.sun, self.planet, PltOpts.DDD.value, res_dir,
                    SubDir.tex_figs.value, 'contact1')
        plt_profile_approx(res_dir, SubDir.tex_figs.value)
        plt_3d(self.sun.x_axis, self.sun.y_axis, self.sun.press,
               self.sun.x_label, self.sun.y_label, 'pressure in MPa',
               'contact_pressure_sun', res_dir, SubDir.tex_figs.value,
               'pressure1')
        plt_energy_ring_on_ring(self, res_dir, SubDir.tex_figs.value, 'energy1')



import glob
import os
import pickle
import shutil

from Constants import SubDir, NpDBs, PrintOpts
from system_functions import make_directory, print_it


def make_infl_mat_db(results_master_folder):
    """Checks if influenc matrix database exists (pickle file). If no,
    creates pickle file with empty dictionary"""
    dir_path = make_directory(results_master_folder,
                              SubDir.infl_mat_db_folder.value)
    db_fname = os.sep.join([dir_path, NpDBs.infl_mat_db.value])
    if not os.path.exists(db_fname):
        with open(db_fname, 'wb') as handle:
            pickle.dump({}, handle, protocol=pickle.HIGHEST_PROTOCOL)
    return db_fname


def cache_influ_mat(ui, influ_mats, res_dir):
    """Checks if inluence matrix is in cache (and database) already. If no,
    influence matrix is added"""
    infl_mat_folder = os.sep.join(
        [os.path.dirname(res_dir), SubDir.infl_mat_db_folder.value])
    infl_data_dict_path = os.sep.join(
        [infl_mat_folder, NpDBs.infl_mat_db.value])

    with open(infl_data_dict_path, 'rb') as handle:
        infl_data_dict = pickle.load(handle)
    for idx, mat in enumerate(influ_mats):
        value = '{}{}'.format(ui['parameter_id'], idx * '-')
        if value in infl_data_dict.values():
            print_it('influence matrix already in cache', PrintOpts.lvl1.value)
        else:
            print_it("caching influence matrix '{}'".format(mat),
                     PrintOpts.lvl1.value)
            infl_data_dict.update({mat: value})
            shutil.copy(os.sep.join([res_dir, SubDir.np_db.value, mat]),
                        os.sep.join([infl_mat_folder, mat]))

    with open(infl_data_dict_path, 'wb') as handle:
        pickle.dump(infl_data_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)


def load_influ_mat(ui, res_dir, number_matrices):
    """Load cached influence matrix and copy it to the current results folder"""
    infl_mat_folder = os.sep.join(
        [os.path.dirname(res_dir), SubDir.infl_mat_db_folder.value])
    infl_data_dict_path = os.sep.join(
        [infl_mat_folder, NpDBs.infl_mat_db.value])

    with open(infl_data_dict_path, 'rb') as handle:
        infl_data_dict = pickle.load(handle)
    key_list = number_matrices * ['empty']
    for i in range(number_matrices):
        for key, value in infl_data_dict.items():
            if '{}{}'.format(ui['parameter_id'], i * '-') == value:
                print_it("loading cached influence matrix '{}'".format(key),
                         PrintOpts.lvl1.value)
                make_directory(res_dir, SubDir.np_db.value)
                shutil.copy(os.sep.join([infl_mat_folder, key]),
                            os.sep.join([res_dir, SubDir.np_db.value, key]))
                key_list[i] = key
    return key_list


def manage_influ_mat_cache(res_dir):
    """Check if the influence matrix cache size exceeds the size limit. If yes,
    remove oldest influence matrix files from cache until size limit is no
    longer exceeded"""
    print_it('clearing influence matrix cache')
    infl_mat_folder = os.sep.join(
        [os.path.dirname(res_dir), SubDir.infl_mat_db_folder.value])
    infl_data_dict_path = os.sep.join(
        [infl_mat_folder, NpDBs.infl_mat_db.value])
    size_cache = 0
    deletable_files = []

    # get size of cache
    for f in sorted(glob.iglob('{}{}*'.format(infl_mat_folder, os.sep)),
                    key=os.path.getctime):
        size_cache += os.path.getsize(f)
        if f != os.sep.join([infl_mat_folder, NpDBs.infl_mat_db.value]):
            deletable_files.extend(f)
    counter = 0

    # remove file from cache if cache size limit is exceeded
    if size_cache / 1024 ** 2 < NpDBs.max_cache_size.value:
        print_it('cache size smaller than cache size limit ({} MB)'.format(
            NpDBs.max_cache_size.value),
                 PrintOpts.lvl1.value)
        print_it('nothing to clear'.format(NpDBs.max_cache_size.value),
                 PrintOpts.lvl1.value)
    else:
        print_it('cache size larger than cache size limit ({} MB)'.format(
            NpDBs.max_cache_size.value),
                 PrintOpts.lvl1.value)
    while size_cache / 1024 ** 2 > NpDBs.max_cache_size.value:
        print_it(
            'deleting influence matrix {}'.format(deletable_files[counter]),
            PrintOpts.lvl1.value)
        os.remove(str(deletable_files[counter]))
        with open(infl_data_dict_path, 'rb') as handle:
            infl_data_dict = pickle.load(handle)
        del infl_data_dict[os.path.basename(str(deletable_files[counter]))]
        with open(infl_data_dict_path, 'wb') as handle:
            pickle.dump(infl_data_dict, handle,
                        protocol=pickle.HIGHEST_PROTOCOL)

        size_cache = 0
        for f in sorted(glob.iglob('{}{}*'.format(infl_mat_folder, os.sep)),
                        key=os.path.getctime):
            size_cache += os.path.getsize(f)
            if f != NpDBs.infl_mat_db.value:
                deletable_files.append(f)
        counter += 1



"""

Python init file for tribology package

"""

__all__ = ['tribology',
           'boundary_element',
           'constants',
           'data_import',
           'dowson_hamrock',
           'hertz',
           'lubrication',
           'roller_bearings']

from tribology.tribology import *
from tribology.boundary_element import *
from tribology.constants import *
from tribology.data_import import *
from tribology.dowson_hamrock import *
from tribology.hertz import *
from tribology.lubrication import *
from tribology.roller_bearings import *

import sys
packpath = os.sep.join(os.path.realpath(__file__).split(os.sep)[:-1])
sys.path.insert(0, packpath)

subs_dirs = os.walk(packpath)
for sub_dir in subs_dirs:
    if not sub_dir[0].endswith('__pycache__'):
        sys.path.insert(0, sub_dir[0])

pass


import copy
import math
import re

import numpy as np

from BluPrintRollBear import AxialRollBear
from Constants import PltOpts, SubDir, MasterDir, UnitTex, Unit, PrintOpts, \
    PreSol
from cartesian_plot_functions import plt_profile, plt_contact, plt_3d, \
    plt_2d_scatt_line, plt_2d_2y_ax, plt_profile_approx
from hertz_equations import hertz_displ
from influ_matrix_management import cache_influ_mat, load_influ_mat
from solve_half_space import solve_half_space, pre_solve_half_space
from system_functions import print_it, to_preci


class AxialThrustBear(AxialRollBear):
    """Global bearing data"""

    def __init__(self, num_rollers, mean_diameter, roller, ring1, ring2,
                 init_force, bearing_name='bearing'):
        super().__init__(bearing_name, num_rollers, roller, ring1, ring2,
                         init_force, mean_diameter)

        self.press_zone_len = None
        self.overrolling_duration = None
        self.overroll_t_incr = None
        self.non_overroll_t = None
        self.e_akin = None
        self.uniq_norm_forces = None
        self.eff_rot_velocity = None
        self.influ_mat_db_1 = None

    def calc_load_distribution(self, ui=None, res_dir=None):
        """Calculate load distribution"""
        print_it("calculating 1 load distribution")
        self.roller.norm_forces = \
            np.multiply(np.ones(self.num_rollers),
                        self.global_force / self.num_rollers)
        self.uniq_norm_forces = self.roller.norm_forces[0]

    def get_grid_size(self, ui=None, res_dir=None):
        """Determine grid size by running (quick) simulation with simplified
        contact bodies"""
        print_it('determining grid size', PrintOpts.lvl1.value)
        self.roller.simple_clone()
        self.roller.clone.make_profile(PreSol.res_x.value,
                                       PreSol.res_y.value, self.init_force)
        self.ring1.simple_clone()
        self.ring1.clone.make_slave_to(self.roller.clone)

        init_displ = \
            hertz_displ(self.roller.clone.e, self.ring1.clone.e,
                        self.roller.clone.ny, self.ring1.clone.ny,
                        self.roller.clone.r_hertz_x,
                        self.roller.clone.r_hertz_y,
                        self.ring1.clone.r_hertz_x,
                        self.ring1.clone.r_hertz_y, self.roller.norm_forces[0])
        contact_delta = 1
        contact_width = 0.05
        while contact_delta != 0:
            self.roller.clone.make_profile(self.roller.clone.res_x,
                                           self.roller.clone.res_y,
                                           self.init_force,
                                           contact_width=contact_width)
            self.ring1.clone.make_slave_to(self.roller.clone)

            pressure, init_displ = \
                pre_solve_half_space(self.roller.clone.profile,
                                     self.ring1.clone.profile,
                                     self.roller.clone.x_axis,
                                     self.roller.clone.y_axis,
                                     self.roller.clone.res_x,
                                     self.roller.clone.res_y,
                                     self.roller.clone.delta_x,
                                     self.roller.clone.delta_y,
                                     self.roller.clone.e,
                                     self.ring1.clone.e,
                                     self.roller.clone.ny,
                                     self.ring1.clone.ny,
                                     self.roller.norm_forces[0],
                                     init_displ=init_displ)

            pressure_elements = sum(
                pressure[math.floor(self.roller.clone.res_y / 2), :] > 0)
            contact_delta = pressure_elements + 2 - self.roller.clone.res_y
            contact_width += np.sign(contact_delta) * contact_width / 25

        self.roller.make_profile(self.roller.res_x, self.roller.res_y,
                                 self.init_force, contact_width=contact_width)
        self.ring1.make_slave_to(self.roller)
        self.ring2.make_slave_to(self.roller)
        return init_displ

    def calc_contact_pressure(self, ui=None, res_dir=None):
        """Calculate contact pressure distribution between washer(s) and
        roller"""
        print_it('calculating 1 pressure distribution')
        init_displ = self.get_grid_size(ui, res_dir)
        [self.influ_mat_db_1] = load_influ_mat(ui, res_dir, 1)
        print_it('solving first half space', PrintOpts.lvl1.value)
        self.roller.press, self.influ_mat_db_1 = \
            solve_half_space(self.roller.profile, self.ring1.profile,
                             self.roller.x_axis, self.roller.y_axis,
                             self.roller.res_x, self.roller.res_y,
                             self.roller.delta_x, self.roller.delta_y,
                             self.roller.e, self.ring1.e, self.roller.ny,
                             self.ring1.ny, self.roller.norm_forces[0],
                             res_dir, init_displ=init_displ,
                             influ_mat_db=self.influ_mat_db_1)
        self.roller.max_press = np.amax(self.roller.press, axis=1)
        self.ring1.press = copy.copy(self.roller.press)
        self.ring1.max_press = copy.copy(self.roller.max_press)
        cache_influ_mat(ui, [self.influ_mat_db_1], res_dir)

    def calc_kinematics(self, rot_vel1, rot_vel2, ui=None, res_dir=None):
        """Calculate bearing kinematics based on rotational velocities of
        washer disks"""
        print_it("calculating kinematics")
        self.rot_vel1 = rot_vel1
        self.rot_vel2 = rot_vel2
        self.eff_rot_velocity = self.rot_vel1 - self.rot_vel2
        self.eff_omega = self.eff_rot_velocity / 60
        self.eff_omega_cage = self.eff_omega / 2
        self.roller.eff_omega = self.mean_diameter / self.roller.diameter * \
                                self.eff_omega_cage
        self.footpr_vel = 2 * math.pi * \
                          (self.mean_diameter / 2 + self.roller.x_axis) * \
                          self.eff_omega / 2

        self.overroll_t_incr = self.roller.delta_y / self.footpr_vel
        self.press_zone_len = (self.ring1.press > 0).sum(1) * self.ring1.delta_y
        self.overrolling_duration = np.divide(self.press_zone_len,
                                              self.footpr_vel)
        self.non_overroll_t = np.divide(
            (2 * math.pi * (self.mean_diameter / 2 + self.roller.x_axis) -
             self.num_rollers * self.press_zone_len) / self.num_rollers,
            self.footpr_vel)

        self.ring1.eff_raceway_vel = 2 * math.pi * (
        self.mean_diameter / 2 + self.roller.x_axis) * self.eff_omega
        self.ring2.eff_raceway_vel = 0
        self.roller.actual_raceway_vel = 2 * math.pi * (
        self.mean_diameter / 2 + self.roller.x_axis) * \
                                 self.eff_omega_cage + math.pi * \
                                 self.roller.diameter * \
                                 self.roller.eff_omega
        self.roller.eff_raceway_vel = self.roller.actual_raceway_vel
        self.roller.rel_vel = math.pi * self.roller.diameter * \
                              self.roller.eff_omega - 2 * math.pi * \
                              (self.mean_diameter / 2 + self.roller.x_axis) * \
                              self.eff_omega_cage
        self.roller.slip = np.divide(self.roller.rel_vel, 2 * math.pi *
                                     (self.mean_diameter / 2 +
                                      self.roller.x_axis) *
                                     self.eff_omega_cage)

    def calc_pv(self, ui=None, res_dir=None):
        """Calculate product of local maximum pressure and local maximum
        relative velocity"""
        print_it("calculating pv_rel")
        self.pv = np.absolute(np.multiply(self.ring1.max_press,
                                          self.roller.rel_vel)) / 1000

    def calc_e_akin(self, ui=None, res_dir=None):
        """"Calculate the kinetic friction energy accumulation in W per m^2"""
        print_it("calculating e_a,kin")
        pv_local = np.multiply(self.ring1.press.sum(1), self.roller.rel_vel)
        self.e_akin = np.absolute(np.divide(np.multiply(pv_local,
                                                        self.overroll_t_incr),
                                            self.non_overroll_t)) / 1000

    def plot_it(self, ui=None, res_dir=None):
        """Orchestrate output plot generation"""
        print_it("plotting results and finishing up")
        self.plt_profiles_and_contacts(res_dir)
        self.plt_pol_slip(res_dir)
        self.plt_pres(res_dir)
        self.plt_pv(res_dir)
        self.plt_eakin(res_dir)

    def plt_profiles_and_contacts(self, res_dir):
        """Generate plots of profiles and contacts"""
        plt_profile(self.roller, PltOpts.DD.value, res_dir,
                    SubDir.profiles.value)
        plt_profile(self.roller, PltOpts.DDD.value, res_dir,
                    SubDir.profiles.value)
        plt_profile(self.ring1, PltOpts.DD.value, res_dir,
                    SubDir.profiles.value)
        plt_profile(self.ring1, PltOpts.DDD.value, res_dir,
                    SubDir.profiles.value)
        plt_profile(self.ring2, PltOpts.DD.value, res_dir,
                    SubDir.profiles.value)
        plt_profile(self.ring2, PltOpts.DDD.value, res_dir,
                    SubDir.profiles.value)
        plt_profile_approx(res_dir, SubDir.profiles.value)
        plt_contact(self.roller, self.ring1, PltOpts.DD.value, res_dir,
                    SubDir.contacts.value)
        plt_contact(self.roller, self.ring1, PltOpts.DDD.value, res_dir,
                    SubDir.contacts.value)
        plt_contact(self.roller, self.ring2, PltOpts.DD.value, res_dir,
                    SubDir.contacts.value)
        plt_contact(self.roller, self.ring2, PltOpts.DDD.value, res_dir,
                    SubDir.contacts.value)

    def plt_pol_slip(self, res_dir):
        """Plot roller slip"""
        plt_2d_scatt_line(self.roller.x_axis, self.roller.slip,
                          self.roller.x_axis, self.roller.slip,
                          self.roller.x_label, "roller slip in percent",
                          "roller slip", res_dir,
                          SubDir.energy.value, "roller slip")

    def plt_pres(self, res_dir):
        """Plot contact pressure distribution"""
        plt_3d(self.roller.x_axis, self.roller.y_axis, self.ring1.press,
               self.roller.x_label, self.roller.y_label, 'pressure in MPa',
               'contact pressure', res_dir, SubDir.pressures.value,
               'contact-pressure')

    def plt_pv(self, res_dir):
        """Plot pv"""
        plt_2d_scatt_line(self.roller.x_axis, self.pv, self.roller.x_axis,
                          self.pv, self.roller.x_label,
                          'pv_rel in {}'.format(Unit.pvrel.value),
                          'pv_rel', res_dir, SubDir.energy.value, 'pv_rel')

    def plt_eakin(self, res_dir):
        """Plot kinetic friction energy accumulation"""
        plt_2d_scatt_line(self.roller.x_axis, self.e_akin, self.roller.x_axis,
                          self.e_akin, self.roller.x_label,
                          'e_akin in W $m^{-2}$', 'e_akin', res_dir,
                          SubDir.energy.value, 'e_akin')
        plt_2d_2y_ax(self.roller.x_axis, self.e_akin, self.roller.x_axis,
                     self.pv, self.roller.x_label, 'e_akin vs pv_rel',
                     'e_akin in {}'.format(Unit.eakin.value),
                     'pv_rel in {}'.format(Unit.pvrel.value), res_dir,
                     SubDir.energy.value, 'e_akin_vs_pv_rel')

    def generate_latex_output(self, calc_spec_tex_file_handle, sim, ui=None,
                              res_dir=None):
        """Generate calculation-specific part of the LaTeX output file"""
        mean_press = np.mean(self.ring1.press[:, :][self.ring1.press[:, :] > 0])
        numeric_output_data = [
            ('pv_rel, max.', to_preci(np.amax(self.pv), 4),
             UnitTex.pvrel.value, 'unverified'),
            ('e_akin, max.', to_preci(np.amax(self.e_akin), 4),
             UnitTex.eakin.value, 'unverified'),
            ('pressure, max.', to_preci(np.amax(self.ring1.press[:, :]), 4),
             UnitTex.pressure.value, 'unverified'),
            ('pressure, av.', to_preci(mean_press, 4),
             UnitTex.pressure.value, 'unverified'),
            ('slip, max.', to_preci(np.amax(self.roller.slip) * 100, 4),
             '\%', 'unverified'),
            ('normal force, max.', to_preci(max(self.roller.norm_forces), 4),
             UnitTex.force.value, 'unverified'),
            ('contact area', to_preci(self.roller.get_area(self.ring1.press),
                                      4),
             UnitTex.area.value, 'unverified')]

        scale_factor_ir = 0.5 + 0.1 * (self.rot_vel1 != 0)
        scale_factor_or = 0.5 + 0.1 * (self.rot_vel2 != 0)

        table_calc_summary = []
        for key, value, unit, status in sorted(numeric_output_data):
            table_calc_summary.append(
                (re.sub('_', '\_', key), value, unit, status)
            )

        latex_variables = {'contact_plot1': 'figures{}contact1.png'.format('/'),
                           'pressure_plot1':
                               'figures{}pressure1.png'.format('/'),
                           'slip_plot1': 'figures{}slip1.png'.format('/'),
                           'energy_plot1': 'figures{}energy1.png'.format('/'),
                           'scale_factor_ir': scale_factor_ir,
                           'scale_factor_or': scale_factor_or,
                           'table_calc_summary': table_calc_summary}

        template_calc_specific = sim.latex_jinja_env.get_template(
            MasterDir.latex_templates.value +
            '/' +
            'template-axial-thrust-bearing.tex')
        with open(calc_spec_tex_file_handle, 'w') as f:
            f.write(template_calc_specific.render(latex_variables))

    def generate_latex_figures(self, ui=None, res_dir=None):
        """Generate calculation-specific figures for LaTeX report"""
        plt_profile_approx(res_dir, SubDir.tex_figs.value)
        plt_contact(self.roller, self.ring1, PltOpts.DDD.value, res_dir,
                    SubDir.tex_figs.value, 'contact1')
        plt_3d(self.roller.x_axis, self.roller.y_axis, self.ring1.press[:, :],
               self.roller.x_label, self.roller.y_label,
               'pressure in MPa', 'pressure at highest normal force', res_dir,
               SubDir.tex_figs.value, 'pressure1')
        plt_2d_scatt_line(self.roller.x_axis, self.roller.slip,
                          self.roller.x_axis, self.roller.slip,
                          self.roller.x_label, "roller slip in percent (0-1)",
                          "roller slip", res_dir,
                          SubDir.tex_figs.value, "slip1")
        plt_2d_2y_ax(self.roller.x_axis, self.e_akin, self.roller.x_axis,
                     self.pv, self.roller.x_label, 'e_akin vs pv_rel',
                     'e_akin in {}'.format(Unit.eakin.value),
                     'pv_rel in {}'.format(Unit.pvrel.value), res_dir,
                     SubDir.tex_figs.value, 'energy1')



import getpass
import math
import os
import platform
import re
import shutil
import subprocess
import sys

import jinja2
import matplotlib as mpl
import numexpr
import numpy as np
import scipy as sc

from Constants import SubDir, MasterDir, PrintOpts, TexTempl
from system_functions import make_directory, print_it, get_p3can_version


def generate_latex_output(simulation, tribo_system, ui=None, res_dir=None):
    """Orchestrate the LaTeX report generation. Load templates, set-up jinja2
    environment, define jinja2 variables, write LaTeX code to output file and
    trigger call to pdflatex"""
    print_it('generating simulation report by calling pdflatex')
    tex_file_path = make_directory(res_dir, SubDir.latex_files.value)
    make_directory(res_dir, SubDir.tex_figs.value)

    tex_filename = 'p3can_simulation_report.tex'
    tex_file_handle = tex_file_path + '/' + tex_filename
    calc_spec_tex_filename = 'p3can_simulation_report_calc_specific.tex'
    calc_spec_tex_file_handle = tex_file_path + '/' + calc_spec_tex_filename

    simulation.latex_jinja_env = jinja2.Environment(
        block_start_string='\BLOCK{',
        block_end_string='}',
        variable_start_string='\VAR{',
        variable_end_string='}',
        comment_start_string='\#{',
        comment_end_string='}',
        line_statement_prefix='%%',
        line_comment_prefix='%#',
        trim_blocks=True,
        autoescape=False,
        loader=jinja2.FileSystemLoader(os.path.abspath('.'))
    )

    template_report = simulation.latex_jinja_env.get_template(
        '/'.join([MasterDir.latex_templates.value,
                  TexTempl.Generic.value]).replace(os.getcwd(), ''))
    print_it('generating calculation-specific latex output',
             PrintOpts.lvl1.value)
    tribo_system.generate_latex_output(calc_spec_tex_file_handle, simulation)
    print_it('generating report figures', PrintOpts.lvl1.value)
    tribo_system.generate_latex_figures(res_dir=res_dir)

    with open(calc_spec_tex_file_handle, 'r') as calc_specific_latex_file:
        calc_specific_output = calc_specific_latex_file.read()

    latex_variables = {'section1': 'Long Form',
                       'section2': 'Short Form',
                       'sim_name': simulation.simulation_name,
                       'p3can_version': get_p3can_version(PrintOpts.lvl1.value),
                       'user_name': getpass.getuser(),
                       'python_version': re.sub('_', '\_',
                                                re.sub('#', '\#', sys.version)),
                       'platform_version': re.sub('_', '\_',
                                                  re.sub('#', '\#',
                                                         platform.version())),
                       'mpl_version': mpl.__version__,
                       'numexpr_version': numexpr.__version__,
                       'np_version': np.__version__,
                       'sc_version': sc.__version__,
                       'machine': re.sub('_', '\_', platform.machine()),
                       'sys_platform': platform.system(),
                       'run_time': (
                       str(math.ceil(simulation.get_time_elapsed())) + ' s'),
                       'user_input': generate_user_input_table(ui),
                       'calc_specific_output': str(calc_specific_output)}
    with open(tex_file_handle, 'w') as f:
        f.write(template_report.render(latex_variables))
    latex_it(tex_filename, res_dir)


def latex_it(tex_file, res_dir):
    """Compile tex_file by calling pdflatex. Raise compiler error in case
    something goes wrong. Delete LaTeX files after successful compilation
    (except for pdf file)"""
    print_it('calling pdflatex two times', PrintOpts.lvl1.value)
    pdf_filename = 'p3can_simulation_report.pdf'
    os.chdir(str(res_dir + os.sep + 'latex-files'))
    try:
        for _ in range(2):
            cmd = ['pdflatex', '-interaction', 'nonstopmode', tex_file]
            proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            out, err = proc.communicate()
            raise_pdflatex_compiler_error(proc, pdf_filename, out, err, cmd)

        move_to_directory = re.sub("latex-files", "", os.getcwd())
        os.rename((os.sep.join([os.getcwd(), 'p3can_simulation_report.pdf'])),
                  (os.sep.join(
                      [move_to_directory, 'p3can_simulation_report.pdf'])))

        print_it('deleting latex files', PrintOpts.lvl1.value)
        os.chdir('..{}'.format(os.sep))
        shutil.rmtree('latex-files')
        os.chdir('..{}'.format(os.sep))
        print_it('{}report written to: {}'
                 .format(PrintOpts.lvl1.value, os.sep.join(
            (os.getcwd().strip("Results{}".format(os.sep)), res_dir,
             'p3can_simulation_report.pdf'))))

    except OSError as e:
        if e.errno == os.errno.ENOENT:
            print_it('no pdflatex installation found', PrintOpts.lvl1.value)
            print_it('proceeding without generating report',
                     PrintOpts.lvl1.value)
            print_it(
                'check out https://www.latex-project.org/ '
                'if you want to install latex',
                PrintOpts.lvl1.value)
        else:
            raise


def raise_pdflatex_compiler_error(proc, pdf_filename, out, err, cmd):
    """Print LaTeX compiler error message to console"""
    retcode = proc.returncode
    if not retcode == 0:
        os.unlink(pdf_filename)
        print('{}pdflatex encountered an error. here is what it says:\n'.format(
            PrintOpts.lvl1.value),
              out.decode('ascii'), err.decode('ascii'))
        raise ValueError(
            'Error {} executing command: {}'.format(retcode, ' '.join(cmd)))


def generate_user_input_table(ui):
    """Generate user input table in a format that allows for copy-paste from
    LaTeX pdf to python editor"""
    calculation_input = []
    for key, value in ui.items():
        if key != 'parameter_id':
            calculation_input.append(key)
    calculation_input = sorted(calculation_input)
    table_user_input = []
    for i in calculation_input:
        if isinstance(ui[i], str):
            table_user_input.append((i, ("'" + str(ui[i]) + "'")))
        else:
            table_user_input.append((i, str(ui[i])))
    return table_user_input


def get_calc_specific_latex_template(latex_template, simulation):
    """Return the calculation-specific LaTeX template"""
    return simulation.latex_jinja_env.get_template(
        '/'.join([MasterDir.latex_templates.value, latex_template])
            .replace(os.getcwd(), ''))
